# -*- coding: utf-8 -*-
"""
Automatic tests for module ldap0.ldapobject
"""

import os
import unittest
import pickle
import time
from errno import ENOTCONN

# Switch off processing .ldaprc or ldap0.conf before importing _ldap
os.environ['LDAPNOINIT'] = '1'

import _libldap0

import ldap0
import ldap0.schema.util
from ldap0.ldapurl import LDAPUrl
from ldap0.sasl import SaslPasswordAuth
from ldap0.test import SlapdTestCase
from ldap0.ldapobject import LDAPObject, ReconnectLDAPObject
from ldap0.err import NoUniqueEntry
from ldap0.controls.sessiontrack import SessionTrackingControl, SESSION_TRACKING_FORMAT_OID_USERNAME
from ldap0.controls.readentry import PreReadControl, PostReadControl
from ldap0.controls.pagedresults import SimplePagedResultsControl
from ldap0.controls.simple import ProxyAuthzControl, RelaxRulesControl, ManageDSAITControl
from ldap0.controls.libldap import AssertionControl, MatchedValuesControl
from ldap0.extop.whoami import WhoAmIRequest, WhoAmIResponse

LDIF_TEMPLATE = """dn: %(suffix)s
objectClass: organization
o: %(o)s

dn: %(rootdn)s
objectClass: applicationProcess
objectClass: simpleSecurityObject
cn: %(rootcn)s
userPassword: %(rootpw)s

dn: cn=user1,%(suffix)s
objectClass: applicationProcess
objectClass: simpleSecurityObject
cn: user1
userPassword: user1_pw

dn: cn=Foo2,%(suffix)s
objectClass: organizationalRole
cn: Foo2

dn: cn=Foo1,%(suffix)s
objectClass: organizationalRole
cn: Foo1

dn: cn=Foo3,%(suffix)s
objectClass: organizationalRole
cn: Foo3

dn: ou=Container,%(suffix)s
objectClass: organizationalUnit
ou: Container

dn: cn=Foo5,ou=Container,%(suffix)s
objectClass: organizationalRole
cn: Foo5

dn: cn=Foo4,ou=Container,%(suffix)s
objectClass: organizationalRole
cn: Foo4

dn: cn=Foo7,ou=Container,%(suffix)s
objectClass: organizationalRole
cn: Foo7

dn: cn=Foo6,ou=Container,%(suffix)s
objectClass: organizationalRole
cn: Foo6

dn: ou=äöüÄÖUß,ou=Container,%(suffix)s
objectClass: organizationalUnit
ou: äöüÄÖUß

"""


class Test00_LDAPObject(SlapdTestCase):
    """
    test LDAP search operations
    """
    ldap_object_class = LDAPObject
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        super(Test00_LDAPObject, cls).setUpClass()
        # insert some Foo* objects via ldapadd
        cls.server.ldapadd(
            (LDIF_TEMPLATE % {
                'suffix':cls.server.suffix,
                'rootdn':cls.server.root_dn,
                'rootcn':cls.server.root_cn,
                'rootpw':cls.server.root_pw,
                'o': cls.server.suffix.split(',')[0][3:],
            }).encode('utf-8')
        )

    def setUp(self):
        try:
            self._ldap_conn
        except AttributeError:
            # open local LDAP connection
            self._ldap_conn = self._open_ldap_conn()

    def test000_basics(self):
        self.assertEqual(self._ldap_conn.protocol_version, 3)
        self.assertIsInstance(self._ldap_conn.fileno(), int)
        with self.assertRaises(TypeError):
            self.ldap_object_class(0)
        with self.assertRaises(ValueError):
            self.ldap_object_class('ldap.example.com')
        with self.ldap_object_class(LDAPUrl(self.server.ldap_uri)) as l:
            self.assertEqual(
                l.uri,
                LDAPUrl(self.server.ldap_uri).connect_uri()
            )

    def test001_search_subtree(self):
        result = self._ldap_conn.search_s(
            self.server.suffix,
            ldap0.SCOPE_SUBTREE,
            '(cn=Foo*)',
            attrlist=['*'],
        )
        result.sort(key=lambda x: x.dn_s)
        self.assertEqual(
            [(lr.dn_s, lr.entry_as) for lr in result],
            [
                (
                    'cn=Foo%d,%s' % (i, self.server.suffix),
                    {'cn': [b'Foo%d' % i], 'objectClass': [b'organizationalRole']}
                )
                for i in range(1, 4)
            ] + [
                (
                    'cn=Foo%d,ou=Container,%s' % (i, self.server.suffix),
                    {'cn': [b'Foo%d' % i], 'objectClass': [b'organizationalRole']}
                )
                for i in range(4, 8)
            ]
        )

    def test002_search_onelevel(self):
        result = self._ldap_conn.search_s(
            self.server.suffix,
            ldap0.SCOPE_ONELEVEL,
            '(cn=Foo*)',
            ['*'],
        )
        result.sort(key=lambda x: x.dn_s)
        self.assertEqual(
            [(lr.dn_s, lr.entry_as) for lr in result],
            [
                (
                    'cn=Foo1,'+self.server.suffix,
                    {'cn': [b'Foo1'], 'objectClass': [b'organizationalRole']}
                ),
                (
                    'cn=Foo2,'+self.server.suffix,
                    {'cn': [b'Foo2'], 'objectClass': [b'organizationalRole']}
                ),
                (
                    'cn=Foo3,'+self.server.suffix,
                    {'cn': [b'Foo3'], 'objectClass': [b'organizationalRole']}
                ),
            ]
        )

    def test003_search_oneattr(self):
        result = self._ldap_conn.search_s(
            self.server.suffix,
            ldap0.SCOPE_SUBTREE,
            '(cn=Foo4)',
            ['cn'],
        )
        result.sort()
        self.assertEqual(
            [(lr.dn_s, lr.entry_as) for lr in result],
            [('cn=Foo4,ou=Container,'+self.server.suffix, {'cn': [b'Foo4']})]
        )

    def test004_errno107(self):
        with self.ldap_object_class('ldap://127.0.0.1:42') as l:
            try:
                ldap_res = l.simple_bind_s('', b'')
            except ldap0.SERVER_DOWN as ldap_err:
                err_no = ldap_err.args[0]['errno']
                if err_no != ENOTCONN:
                    self.fail("expected err_no=%d, got %d" % (ENOTCONN, err_no))
                info = ldap_err.args[0]['info']
                info_expected = os.strerror(ENOTCONN).encode('ascii')
                if info != info_expected:
                    self.fail("expected info=%r, got %r" % (info_expected, info))
            else:
                self.fail("expected SERVER_DOWN, got %r" % ldap_res)

    def test005_invalid_credentials(self):
        with self.ldap_object_class(self.server.ldap_uri) as l:
            try:
                ldap_res = l.simple_bind_s(
                    self.server.root_dn,
                    (self.server.root_pw+'wrong').encode('utf-8'),
                )
            except ldap0.INVALID_CREDENTIALS:
                pass
            else:
                self.fail("expected INVALID_CREDENTIALS, got %r" % ldap_res)

    def test006_sasl_external(self):
        with self.ldap_object_class(self.server.ldapi_uri) as l:
            l.sasl_non_interactive_bind_s('EXTERNAL')
            self.assertEqual(l.whoami_s(), 'dn:'+self.server.root_dn.lower())
        authz_id = 'dn:cn=Foo2,%s' % (self.server.suffix)
        with self.ldap_object_class(self.server.ldapi_uri) as l:
            l.sasl_non_interactive_bind_s('EXTERNAL', authz_id=authz_id)
            self.assertEqual(l.whoami_s(), authz_id.lower())

    def test006_sasl_password_mechs(self):
        authz_id = 'dn:cn=user1,%s' % (self.server.suffix)
        for sasl_mech, sasl_ssf in (
                ('PLAIN', None),
                ('CRAM-MD5', None),
                ('DIGEST-MD5', 128),
#                ('SCRAM-SHA-1', None),
            ):
            with self.ldap_object_class(self.server.ldapi_uri) as l:
                l.sasl_interactive_bind_s(
                    sasl_mech,
                    SaslPasswordAuth('user1', b'user1_pw'),
                )
                if hasattr(ldap0, 'OPT_X_SASL_SSF') and sasl_ssf is not None:
                    self.assertEqual(l.get_option(ldap0.OPT_X_SASL_SSF), sasl_ssf)
                self.assertEqual(l.whoami_s(), authz_id.lower())

    def test006_sasl_password_mechs_bad(self):
        for sasl_mech in (
                'PLAIN',
                'CRAM-MD5',
                'DIGEST-MD5',
#                'SCRAM-SHA-1',
            ):
            with self.ldap_object_class(self.server.ldapi_uri) as l:
                with self.assertRaises(ldap0.INVALID_CREDENTIALS):
                    l.sasl_interactive_bind_s(
                        sasl_mech,
                        SaslPasswordAuth('user1', b'badpassword'),
                    )

    @unittest.skip('Too timing dependent')
    def test007_timeout(self):
        test_timeout = 0.00001
        with self.ldap_object_class(self.server.ldap_uri) as l:
            with self.assertRaises(ldap0.TIMEOUT):
                msg_id = l.search(
                    self.server.suffix,
                    ldap0.SCOPE_SUBTREE,
                    '(objectClass=*)',
                )
                l.result(msg_id, all_results=ldap0.MSG_ALL, timeout=test_timeout)
            l.abandon(msg_id)
            with self.assertRaises(ldap0.TIMEOUT):
                l.search_s(
                    self.server.suffix,
                    ldap0.SCOPE_SUBTREE,
                    '(objectClass=*)',
                    timeout=test_timeout,
                )

    def test008_sessiontrack_control(self):
        self._ldap_conn.search_s(
            '',
            ldap0.SCOPE_BASE,
            req_ctrls=[
                SessionTrackingControl(
                    '192.0.2.1',
                    'app.example.com',
                    SESSION_TRACKING_FORMAT_OID_USERNAME,
                    'bloggs'
                )
            ],
        )

    def test009_readentry_control(self):
        attr_list = sorted([
            'entryUUID',
            'entryDN',
            'createTimestamp',
            'modifyTimestamp',
        ])
        new_dn = 'cn=Anna Blume,'+self.server.suffix
        # add a new entry and return operational attributes with post-read control
        add_result = self._ldap_conn.add_s(
            new_dn,
            {
                'objectClass': [b'applicationProcess'],
                'cn': [b'Anna Blume'],
            },
            req_ctrls=[
                PostReadControl(
                    criticality=True,
                    attrList=attr_list,
                )
            ],
        )
        self.assertEqual(add_result.ctrls[0].res.dn_s, new_dn)
        self.assertEqual(add_result.ctrls[0].res.entry_as['entryDN'][0].decode('utf-8'), new_dn)
        self.assertEqual(len(add_result.ctrls[0].res.entry_as), len(attr_list))
        self.assertEqual(sorted(add_result.ctrls[0].res.entry_as.keys()), attr_list)
        # modify existing entry and return old operational attributes with pre-read control
        modify_result = self._ldap_conn.modify_s(
            new_dn,
            [
                (ldap0.MOD_ADD, b'description', [b'testing...']),
            ],
            req_ctrls=[
                PreReadControl(
                    criticality=True,
                    attrList=attr_list,
                )
            ],
        )
        self.assertEqual(add_result.ctrls[0].res, modify_result.ctrls[0].res)

    def test010_pagedresults(self):
        expected_ldap_result = []
        expected_ldap_result = self._ldap_conn.search_s(
            self.server.suffix,
            ldap0.SCOPE_SUBTREE,
            attrlist=['1.1'],
        )
        expected_ldap_result.sort(key=lambda x: x.dn_s)
        expected_num_results = len(expected_ldap_result)
        self.assertGreaterEqual(expected_num_results, 10)
        page_size = 3
        expected_num_pages = expected_num_results // page_size + \
                            (expected_num_results % page_size > 0)
        paged_results_control = SimplePagedResultsControl(
            True,
            size=page_size,
            cookie=b''
        )
        result_pages = 0
        all_paged_results = []
        # do the paged search as long as a cookie is received in response control
        paging_continues = True
        while paging_continues:
            msg_id = self._ldap_conn.search(
                self.server.suffix,
                ldap0.SCOPE_SUBTREE,
                attrlist=['1.1'],
                req_ctrls=[paged_results_control],
            )
            ldap_results = list(self._ldap_conn.results(msg_id))
#            self.assertEqual(len(ldap_results), page_size+1)
            for res in ldap_results:
                self.assertEqual(res.msgid, msg_id)
                if res.rtype == ldap0.RES_SEARCH_ENTRY and res.rdata:
                    all_paged_results.extend(res.rdata)
                elif res.rtype == ldap0.RES_SEARCH_RESULT:
                    result_pages += 1
                    if not res.ctrls or not res.ctrls[0].cookie:
                        paging_continues = False
                    self.assertEqual(res.ctrls[0].controlType, SimplePagedResultsControl.controlType)
                    # Copy cookie from response control to request control
                    paged_results_control.cookie = res.ctrls[0].cookie
        self.assertEqual(result_pages, expected_num_pages)
        self.assertEqual(sorted(all_paged_results, key=lambda x: x.dn_s), expected_ldap_result)

    def test011_add_dict(self):
        new_dn = 'cn=Kurt Schwitters,'+self.server.suffix
        new_entry = {
            'objectClass': [b'applicationProcess'],
            'cn': [b'Kurt Schwitters'],
            'description': [b'Dada!'],
        }
        self._ldap_conn.add_s(new_dn, new_entry)
        read = self._ldap_conn.read_s(new_dn, attrlist=new_entry.keys())
        self.assertEqual(read.entry_as, new_entry)
        self._ldap_conn.delete_s(new_dn)

    def test012_results(self):
        msg_id = self._ldap_conn.search(
            self.server.suffix,
            ldap0.SCOPE_SUBTREE,
            '(cn=Foo*)',
            attrlist=['*'],
        )
        result = []
        for ldap_res in self._ldap_conn.results(msg_id):
            for search_res in ldap_res.rdata:
                result.append((search_res.dn_s, search_res.entry_as))
        result.sort()
        self.assertEqual(
            result,
            [
                (
                    ('cn=Foo%d,%s' % (i, self.server.suffix)),
                    {'cn': [b'Foo%d' % i], 'objectClass': [b'organizationalRole']}
                )
                for i in range(1, 4)
            ] + [
                (
                    ('cn=Foo%d,ou=Container,%s' % (i, self.server.suffix)),
                    {'cn': [b'Foo%d' % i], 'objectClass': [b'organizationalRole']}
                )
                for i in range(4, 8)
            ]
        )

    def test013_search_cache(self):
        with self.ldap_object_class(
            self.server.ldapi_uri,
            trace_level=self.trace_level,
            cache_ttl=10.0,
        ) as l:
            self.assertEqual(l._cache_ttl, 10.0)
            l.sasl_non_interactive_bind_s('EXTERNAL')
            self.assertEqual(l.whoami_s(), 'dn:'+self.server.root_dn.lower())
            new_dn = 'cn=Cache Tester,'+self.server.suffix
            new_entry = {
                'objectClass': [b'applicationProcess'],
                'cn': [b'Cache Tester'],
            }
            l.add_s(new_dn, new_entry)
            for i in range(5):
                ldap_res = l.search_s(new_dn, ldap0.SCOPE_BASE)[0]
                self.assertEqual(ldap_res.dn_s, new_dn)
                self.assertEqual(ldap_res.entry_as, new_entry)
                self.assertEqual(l._cache.miss_count, 1)
                self.assertEqual(l._cache.hit_count, i)
                self.assertEqual(len(l._cache), 1)
            cache_hits = l._cache.hit_count
            cache_misses = l._cache.miss_count
            l.modify_s(new_dn, [(ldap0.MOD_ADD, b'description', [b'cache test'])])
            l.modify_s(new_dn, [(ldap0.MOD_DELETE, b'description', [b'cache test'])])
            self.assertEqual(l._cache.miss_count, cache_misses)
            self.assertEqual(l._cache.hit_count, cache_hits)
            self.assertEqual(len(l._cache), 0)
            self.assertEqual(l.cache_hit_ratio(), 80.0)
            l.flush_cache()
            self.assertEqual(l._cache.miss_count, 0)
            self.assertEqual(l._cache.hit_count, 0)
            self.assertEqual(len(l._cache), 0)
            self.assertEqual(l.cache_hit_ratio(), 0.0)
            for i in range(5):
                ldap_res = l.search_s(new_dn, ldap0.SCOPE_BASE)[0]
                self.assertEqual(ldap_res.dn_s, new_dn)
                self.assertEqual(ldap_res.entry_as, new_entry)
                self.assertEqual(l._cache.miss_count, 1)
                self.assertEqual(l._cache.hit_count, i)
                self.assertEqual(len(l._cache), 1)
            _ = l.search_s(
                self.server.suffix,
                ldap0.SCOPE_SUBTREE,
                ('(entryDN=%s)' % (new_dn,)),
            )
            self.assertEqual(len(l._cache), 2)
            l.uncache(new_dn)
            self.assertEqual(l._cache.miss_count, 2)
            self.assertEqual(l._cache.hit_count, i+1)
            self.assertEqual(len(l._cache), 0)
            l.flush_cache()
            for i in range(4):
                ldap_res = l.search_s(new_dn, ldap0.SCOPE_BASE, cache_ttl=0.0001)[0]
                self.assertEqual(ldap_res.dn_s, new_dn)
                self.assertEqual(ldap_res.entry_as, new_entry)
                self.assertEqual(l._cache.miss_count, i+1)
                self.assertEqual(l._cache.hit_count, 0)
                self.assertEqual(len(l._cache), 1)
                time.sleep(0.1)

    def test014_extop_whoami(self):
        with self.ldap_object_class(self.server.ldapi_uri) as l:
            l.sasl_non_interactive_bind_s('EXTERNAL')
            wai = l.extop_s(
                WhoAmIRequest(),
                extop_resp_class=WhoAmIResponse,
            )
            self.assertEqual(wai.responseValue, 'dn:'+self.server.root_dn.lower())
            self.assertEqual(l.whoami_s(), 'dn:'+self.server.root_dn.lower())
        authz_id = 'dn:cn=Foo2,%s' % (self.server.suffix)
        with self.ldap_object_class(self.server.ldapi_uri) as l:
            l.sasl_non_interactive_bind_s('EXTERNAL')
            self.assertEqual(
                l.whoami_s([ProxyAuthzControl(authzId='')]),
                '',
            )
            self.assertEqual(
                l.whoami_s(
                    [ProxyAuthzControl(authzId=authz_id)]
                ),
                authz_id.lower(),
            )
        with self.ldap_object_class(self.server.ldapi_uri) as l:
            l.simple_bind_s('', b'')
            wai = l.extop_s(
                WhoAmIRequest(),
                extop_resp_class=WhoAmIResponse,
            )
            self.assertEqual(wai.responseValue, '')
            self.assertEqual(l.whoami_s(), '')

    def test015_passwd(self):
        # first, create a user to change password on
        dn = "cn=PasswordTest," + self.server.suffix
        self._ldap_conn.add_s(
            dn,
            {
                'objectClass': [b'person'],
                'sn': [b'PasswordTest'],
                'cn': [b'PasswordTest'],
                'userPassword': [b'initial'],
            }
        )
        # try changing password with a wrong old-pw
        with self.assertRaises(ldap0.UNWILLING_TO_PERFORM):
            res = self._ldap_conn.passwd_s(dn, b'bogus', b'ignored')
        # try changing password with a correct old-pw
        new_pwd = self._ldap_conn.passwd_s(dn, b'initial', b'changed')
        self.assertEqual(new_pwd, None)
        with self.ldap_object_class(self.server.ldapi_uri) as l:
            l.simple_bind_s(dn, 'changed')
        # let server set new random password
        new_pwd = self._ldap_conn.passwd_s(dn, b'changed', None)
        self.assertIsInstance(new_pwd, bytes)
        self.assertTrue(len(new_pwd)>0)
        with self.ldap_object_class(self.server.ldapi_uri) as l:
            l.simple_bind_s(dn, new_pwd)

    def test016_cancel(self):
        try:
            res = self._ldap_conn.cancel_s(999999999)
        except ldap0.NO_SUCH_OPERATION:
            pass
        else:
            self.fail("expected ldap0.NO_SUCH_OPERATION to be raised")
        # start a "long-lasting" search
        msg_id = self._ldap_conn.search(
            self.server.suffix,
            ldap0.SCOPE_SUBTREE,
            '(objectClass=*)',
            attrlist=['*', '+'],
        )
        time.sleep(1.0)
        try:
            res = self._ldap_conn.cancel_s(msg_id)
        except ldap0.NO_SUCH_OPERATION:
            pass
        else:
            self.fail("expected ldap0.NO_SUCH_OPERATION to be raised")

    def test017_relax_rules(self):
        dn = "cn=RelaxRules," + self.server.suffix
        entry = {
            'objectClass': [b'person'],
            'sn': [b'RelaxRules'],
            'cn': [b'RelaxRules'],
            'createTimestamp': [b'19680208115125Z'],
        }
        try:
            self._ldap_conn.add_s(dn, entry)
        except ldap0.CONSTRAINT_VIOLATION:
            pass
        else:
            self.fail("should have raised ldap0.CONSTRAINT_VIOLATION")
        self._ldap_conn.add_s(
            dn,
            entry,
            req_ctrls=[RelaxRulesControl()],
        )

    def test018_referrals(self):
        opt_ref_old = self._ldap_conn.get_option(ldap0.OPT_REFERRALS)
        self._ldap_conn.set_option(ldap0.OPT_REFERRALS, False)
        dn = "dc=relax-rules," + self.server.suffix
        entry = {
            'objectClass': [b'dcObject', b'referral'],
            'dc': [b'relax-rules'],
            'ref': [b'ldap://ldap.uninett.no/dc=uninett,dc=no??one?objectClass=*)'],
        }
        self._ldap_conn.add_s(dn, entry)
        try:
            ldap_result = self._ldap_conn.search_s(
                dn,
                ldap0.SCOPE_BASE,
                attrlist=['*', 'ref']
            )
        except ldap0.REFERRAL as ldap_ref:
            self.assertEqual(
                ldap_ref.args[0]['desc'],
                b'Referral'
            )
            self.assertEqual(
                ldap_ref.args[0]['matched'],
                dn.encode('utf-8')
            )
            self.assertEqual(
                ldap_ref.args[0]['info'],
                b'Referral:\nldap://ldap.uninett.no/dc=uninett,dc=no??one?objectClass=*)'
            )
        else:
            self.fail('should have raised ldap0.REFERRAL')
        ldap_result = self._ldap_conn.search_s(
            dn,
            ldap0.SCOPE_BASE,
            attrlist=['*', 'ref'],
            req_ctrls=[ManageDSAITControl()],
        )
        self.assertEqual(ldap_result[0].dn_s, dn)
        self.assertEqual(ldap_result[0].entry_as, entry)
        ldap_result = self._ldap_conn.search_s(
            self.server.suffix,
            ldap0.SCOPE_SUBTREE,
            '(dc=relax-rules)',
            attrlist=['*', 'ref'],
        )
        self.assertEqual(
            ldap_result[0].ref_url_strings,
            ['ldap://ldap.uninett.no/dc=uninett,dc=no??one?objectClass=*)']
        )
        ldap_result = self._ldap_conn.search_s(
            self.server.suffix,
            ldap0.SCOPE_SUBTREE,
            '(dc=relax-rules)',
            attrlist=['*', 'ref'],
            req_ctrls=[ManageDSAITControl()],
        )
        self.assertEqual(ldap_result[0].dn_s, dn)
        self.assertEqual(ldap_result[0].entry_as, entry)
        self._ldap_conn.set_option(ldap0.OPT_REFERRALS, opt_ref_old)
        # clean up
        self._ldap_conn.delete_s(dn, req_ctrls=[ManageDSAITControl()])
        # end of test018_referrals()

    def test019_assertion_control(self):
        dn = "cn=AssertionControl," + self.server.suffix
        entry = {
            'objectClass': [b'person'],
            'sn': [b'AssertionControl'],
            'cn': [b'AssertionControl'],
        }
        self._ldap_conn.add_s(dn, entry)
        old = self._ldap_conn.read_s(
            dn,
            attrlist=['entryDN', 'modifyTimestamp', 'entryCSN', 'entryUUID'],
        )
        assertion_filter = (
            '(&'
              '(entryCSN={entryCSN})'
              '(entryDN={entryDN})'
              '(entryUUID={entryUUID})'
              '(modifyTimestamp={modifyTimestamp})'
            ')'
        ).format(
            entryDN=old.entry_s['entryDN'][0],
            entryUUID=old.entry_s['entryUUID'][0],
            modifyTimestamp=old.entry_s['modifyTimestamp'][0],
            entryCSN=old.entry_s['entryCSN'][0],
        )
        self._ldap_conn.modify_s(
            dn,
            [(ldap0.MOD_REPLACE, b'description', [b'test modification #1'])],
            req_ctrls=[AssertionControl(criticality=False, filterstr=assertion_filter)],
        )
        with self.assertRaises(ldap0.ASSERTION_FAILED):
            self._ldap_conn.modify_s(
                dn,
                [(ldap0.MOD_REPLACE, b'description', [b'test modification #2'])],
                req_ctrls=[AssertionControl(filterstr=assertion_filter)],
            )

    def test020_matched_values(self):
        dn = "sn=MatchedValues," + self.server.suffix
        entry = {
            'objectClass': [b'person', b'dcObject'],
            'dc': [b'matched-values'],
            'sn': [b'MatchedValues'],
            'cn': [
                b'MatchedValues0',
                b'MatchedValues1',
                b'MatchedValues2',
                b'MatchedValues3',
                b'MatchedValues4',
                b'MatchedValues5',
            ],
        }
        self._ldap_conn.add_s(dn, entry)
        for val in entry['cn']:
            mv = self._ldap_conn.read_s(
                dn,
                attrlist=['objectClass', 'cn'],
                req_ctrls=[
                    MatchedValuesControl(
                        filterstr='((objectClass=person)(cn=%s))' % (val.decode('utf-8'),)
                    ),
                ]
            )
            self.assertEqual(
                mv.entry_as,
                {'cn': [val], 'objectClass': [b'person']}
            )

    def test021_compare(self):
        self.assertTrue(
            self._ldap_conn.compare_s(
                self.server.root_dn,
                'userPassword',
                self.server.root_pw.encode('utf-8'),
            )
        )
        self.assertFalse(
            self._ldap_conn.compare_s(
                self.server.root_dn,
                'userPassword',
                (self.server.root_pw+'1').encode('utf-8'),
            )
        )
        with self.assertRaises(ldap0.NO_SUCH_OBJECT):
            self._ldap_conn.compare_s('cn=foobar', 'cn', b'foobar')
        with self.assertRaises(ldap0.UNDEFINED_TYPE):
            self._ldap_conn.compare_s('cn=foobar', 'foo', b'bar')

    def test022_get_whoami_dn(self):
        self.assertEqual(
            self._ldap_conn.get_whoami_dn(),
            self.server.root_dn,
        )
        # repeated (cached)
        self.assertEqual(
            self._ldap_conn.get_whoami_dn(),
            self.server.root_dn,
        )
        with self.ldap_object_class(self.server.ldapi_uri) as l:
            self.assertEqual(l._whoami_dn, None)
            l.simple_bind_s('', b'')
            self.assertEqual(l.whoami_s(), '')
            self.assertEqual(l.get_whoami_dn(), '')
            bind_dn = 'cn=user1,'+self.server.suffix
            l.simple_bind_s(bind_dn, b'user1_pw')
            self.assertEqual(
                l.get_whoami_dn(),
                bind_dn,
            )

    def test023_find_unique_entry(self):
        ldap_res = self._ldap_conn.find_unique_entry(
            self.server.suffix,
            ldap0.SCOPE_SUBTREE,
            '(cn=user1)',
        )
        self.assertEqual(ldap_res.dn_s, 'cn=user1,'+self.server.suffix)
        self.assertEqual(
            ldap_res.entry_as,
            {
                'cn': [b'user1'],
                'userPassword': [b'user1_pw'],
                'objectClass': [b'applicationProcess', b'simpleSecurityObject'],
            }
        )
        ldap_res = self._ldap_conn.find_unique_entry(
            self.server.suffix,
            ldap0.SCOPE_SUBTREE,
            '(cn=user1)',
            attrlist=['1.1'],
        )
        self.assertEqual(ldap_res.dn_s, 'cn=user1,'+self.server.suffix)
        self.assertEqual(ldap_res.entry_as, {})
        ldap_res = self._ldap_conn.find_unique_entry(
            self.server.suffix,
            ldap0.SCOPE_SUBTREE,
            '(cn=user1)',
            attrlist=['cn', 'userPassword'],
        )
        self.assertEqual(ldap_res.dn_s, 'cn=user1,'+self.server.suffix)
        self.assertEqual(ldap_res.entry_as, {'cn': [b'user1'], 'userPassword': [b'user1_pw']})
        with self.assertRaises(NoUniqueEntry):
            self._ldap_conn.find_unique_entry(
                self.server.suffix,
                ldap0.SCOPE_SUBTREE,
                '(objectClass=foobar)',
            )
        with self.assertRaises(NoUniqueEntry):
            self._ldap_conn.find_unique_entry(
                self.server.suffix,
                ldap0.SCOPE_SUBTREE,
                '(objectClass=*)',
            )

    def test024_umlauts(self):
        result = self._ldap_conn.search_s(
            self.server.suffix,
            ldap0.SCOPE_SUBTREE,
            '(ou=*ö*)',
            attrlist=['ou'],
        )
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].dn_s, 'ou=äöüÄÖUß,ou=Container,%s' % (self.server.suffix))
        self.assertEqual(result[0].entry_s, {'ou': ['äöüÄÖUß']})
        self.assertTrue(
            self._ldap_conn.compare_s(
                'ou=äöüÄÖUß,ou=Container,%s' % (self.server.suffix),
                'ou',
                'äöüÄÖUß'.encode('utf-8'),
            )
        )
        add_dn = 'ou=ßÜÖÄüöä,ou=Container,%s' % (self.server.suffix)
        self._ldap_conn.add_s(
            add_dn,
            {
                'objectClass': [b'organizationalUnit'],
                'ou': ['ßÜÖÄüöä'.encode('utf-8')],
            },
        )
        self.assertTrue(
            self._ldap_conn.compare_s(
                add_dn,
                'ou',
                'ßÜÖÄüöä'.encode('utf-8'),
            )
        )
        self._ldap_conn.delete_s(add_dn)

    def test025_search_subschemasubentry_s(self):
        self.assertEqual(
            self._ldap_conn.search_subschemasubentry_s(),
            'cn=Subschema',
        )
        self.assertEqual(
            self._ldap_conn.search_subschemasubentry_s(dn=''),
            'cn=Subschema',
        )
        self.assertEqual(
            self._ldap_conn.search_subschemasubentry_s(dn=self.server.suffix),
            'cn=Subschema',
        )
        # this will cause fall back reading attribute subschemaSubentry from rootDSE
        self.assertEqual(
            self._ldap_conn.search_subschemasubentry_s('cn=doesnotexist'),
            'cn=Subschema',
        )
        # DN of subschema subentry cannot be determined and will return None
        self.assertEqual(
            self._ldap_conn.search_subschemasubentry_s('cn=doesnotexist', rootdse=False),
            None,
        )

    def test001_add_modify_rename_delete(self):
        dn1 = 'cn=äöüÄÖÜß,'+self.server.suffix
        rdn2 = 'cn=LaLeLu4711'
        dn2 = ','.join((rdn2, self.server.suffix))
        entry = {
            'objectClass': [b'applicationProcess'],
            'cn': ['äöüÄÖÜß'.encode('utf-8')],
        }
        add_result = self._ldap_conn.add_s(dn1, entry)
        self.assertEqual(
            (add_result.rtype, add_result.rdata, add_result.msgid, add_result.ctrls),
            (_libldap0.RES_ADD, [], 2, [])
        )
        with self.assertRaises(_libldap0.ALREADY_EXISTS):
            self._ldap_conn.add_s(dn1, entry)
        modify_result = self._ldap_conn.modify_s(dn1, [(ldap0.MOD_ADD, b'description', ['äöüÄÖÜß'.encode('utf-8')])])
        self.assertEqual(
            (modify_result.rtype, modify_result.rdata, modify_result.msgid, modify_result.ctrls),
            (_libldap0.RES_MODIFY, [], 4, [])
        )
        with self.assertRaises(_libldap0.TYPE_OR_VALUE_EXISTS):
            self._ldap_conn.modify_s(dn1, [(ldap0.MOD_ADD, b'description', ['äöüÄÖÜß'.encode('utf-8')])])
        modify_result = self._ldap_conn.modify_s(dn1, [(ldap0.MOD_DELETE, b'description', ['äöüÄÖÜß'.encode('utf-8')])])
        self.assertEqual(
            (modify_result.rtype, modify_result.rdata, modify_result.msgid, modify_result.ctrls),
            (_libldap0.RES_MODIFY, [], 6, [])
        )
        with self.assertRaises(_libldap0.NO_SUCH_ATTRIBUTE):
            self._ldap_conn.modify_s(dn1, [(ldap0.MOD_DELETE, b'description', ['äöüÄÖÜß'.encode('utf-8')])])
        rename_result = self._ldap_conn.rename_s(
            dn1,
            rdn2,
            newsuperior=self.server.suffix,
        )
        self.assertEqual(
            (rename_result.rtype, rename_result.rdata, rename_result.msgid, rename_result.ctrls),
            (_libldap0.RES_MODRDN, [], 8, [])
        )
        read_result = self._ldap_conn.read_s(dn2)
        self.assertEqual(read_result.entry_s, {'cn': ['LaLeLu4711'], 'objectClass': ['applicationProcess']})
        with self.assertRaises(_libldap0.NO_SUCH_OBJECT):
            self._ldap_conn.rename_s(dn1, rdn2)
        delete_result = self._ldap_conn.delete_s(dn2)
        self.assertEqual(
            (delete_result.rtype, delete_result.rdata, delete_result.msgid, delete_result.ctrls),
            (_libldap0.RES_DELETE, [], 11, [])
        )
        with self.assertRaises(_libldap0.NO_SUCH_OBJECT):
            self._ldap_conn.delete_s(dn2)

    def test026_ensure_entry(self):
        dn1 = 'cn=unicode-täxt,'+self.server.suffix
        dn2 = 'cn=LaLeLu4711,'+self.server.suffix
        entry1 = {
            'objectClass': [b'applicationProcess'],
            'cn': ['unicode-täxt'.encode('utf-8')],
        }
        entry2 = {
            'objectClass': [b'applicationProcess'],
            'cn': [b'LaLeLu4711'],
            'description': [b'Now you see it...'],
        }
        entry3 = {
            'objectClass': [b'applicationProcess', b'pkiUser'],
            'cn': [b'LaLeLu4711'],
            'description': [b'Now you see it...'],
        }
        mod_attr_list = list(entry2.keys())
        # add
        ldap_res = self._ldap_conn.ensure_entry(dn1, entry1)
        self.assertEqual(len(ldap_res), 1)
        self.assertEqual(ldap_res[0].rtype, ldap0.RES_ADD)
        self.assertEqual(
            self._ldap_conn.read_s(dn1, attrlist=mod_attr_list).entry_as,
            entry1
        )
        self._ldap_conn.delete_s(dn1)
        # add
        ldap_res = self._ldap_conn.ensure_entry(
            dn1,
            entry1,
            old_base=self.server.suffix,
            old_filter='(cn=unicode-täxt)',
            old_attrs=list(entry1.keys()),
        )
        self.assertEqual(len(ldap_res), 1)
        self.assertEqual(ldap_res[0].rtype, ldap0.RES_ADD)
        self.assertEqual(
            self._ldap_conn.read_s(dn1, attrlist=mod_attr_list).entry_as,
            entry1
        )
        # nothing to be done
        ldap_res = self._ldap_conn.ensure_entry(dn1, entry1)
        self.assertEqual(ldap_res, [])
        self.assertEqual(
            self._ldap_conn.read_s(dn1, attrlist=mod_attr_list).entry_as,
            entry1
        )
        # delete
        ldap_res = self._ldap_conn.ensure_entry(dn1, None)
        self.assertEqual(len(ldap_res), 1)
        self.assertEqual(ldap_res[0].rtype, ldap0.RES_DELETE)
        with self.assertRaises(_libldap0.NO_SUCH_OBJECT):
            self._ldap_conn.read_s(dn1, attrlist=['1.1'])
        # nothing to be done
        ldap_res = self._ldap_conn.ensure_entry(dn1, None)
        self.assertEqual(ldap_res, [])
        with self.assertRaises(_libldap0.NO_SUCH_OBJECT):
            ldap_res = self._ldap_conn.ensure_entry(dn1, None, del_ignore=False)
        # add
        ldap_res = self._ldap_conn.ensure_entry(dn2, entry2)
        self.assertEqual(len(ldap_res), 1)
        self.assertEqual(ldap_res[0].rtype, ldap0.RES_ADD)
        self.assertEqual(
            self._ldap_conn.read_s(dn2, attrlist=mod_attr_list).entry_as,
            entry2
        )
        # modify
        ldap_res = self._ldap_conn.ensure_entry(dn2, entry3)
        self.assertEqual(len(ldap_res), 1)
        self.assertEqual(ldap_res[0].rtype, ldap0.RES_MODIFY)
        self.assertEqual(
            self._ldap_conn.read_s(dn2, attrlist=mod_attr_list).entry_as,
            entry3
        )
        # modify without reading old entry because it's provided as argument
        ldap_res = self._ldap_conn.ensure_entry(dn2, entry2, old_entry=entry3)
        self.assertEqual(len(ldap_res), 1)
        self.assertEqual(ldap_res[0].rtype, ldap0.RES_MODIFY)
        self.assertEqual(
            self._ldap_conn.read_s(dn2, attrlist=mod_attr_list).entry_as,
            entry2
        )
        # rename and modify
        ldap_res = self._ldap_conn.ensure_entry(
            dn1,
            entry1,
            old_base=dn2,
            old_filter='(objectClass=applicationProcess)',
        )
        self.assertEqual(len(ldap_res), 2)
        self.assertEqual(ldap_res[0].rtype, ldap0.RES_MODRDN)
        self.assertEqual(ldap_res[1].rtype, ldap0.RES_MODIFY)
        self.assertEqual(
            self._ldap_conn.read_s(dn1, attrlist=mod_attr_list).entry_as,
            entry1
        )
        # clean up
        self._ldap_conn.delete_s(dn1)
        # now some errors....
        with self.assertRaises(ValueError) as ctx:
            self._ldap_conn.ensure_entry(None, None)
        self.assertEqual(str(ctx.exception), 'Expected either dn or old_base to be not None!')

    def test027_res_call_args(self):
        with self.ldap_object_class(self.server.ldapi_uri) as l:
            l.res_call_args = True
            bind_res = l.simple_bind_s('', b'')
            self.assertEqual(l.whoami_s(), '')
            func, args, kwargs = bind_res.call_args
            self.assertEqual(func.__name__, self.ldap_object_class.simple_bind.__name__)
            self.assertEqual(args, (b'', b'', None))
            self.assertEqual(kwargs, {})

    def test028_start_tls(self):
        with self.assertRaises(ldap0.CONNECT_ERROR):
            with self.ldap_object_class(self.server.ldap_uri) as l:
                l.start_tls_s()
        with self.ldap_object_class(self.server.ldap_uri) as l:
            if hasattr(ldap0, 'OPT_SOCKET_BIND_ADDRESSES'):
                l.set_option(ldap0.OPT_SOCKET_BIND_ADDRESSES, b'127.0.0.1')
                self.assertEqual(l.get_option(ldap0.OPT_SOCKET_BIND_ADDRESSES), b'127.0.0.1')
            if hasattr(ldap0, 'OPT_X_TLS_REQUIRE_SAN'):
                l.set_option(ldap0.OPT_X_TLS_REQUIRE_SAN, ldap0.OPT_X_TLS_DEMAND)
                self.assertEqual(l.get_option(ldap0.OPT_X_TLS_REQUIRE_SAN), ldap0.OPT_X_TLS_DEMAND)
            if hasattr(ldap0, 'OPT_X_TLS_PEERKEY_HASH'):
                l.set_option(ldap0.OPT_X_TLS_PEERKEY_HASH, b'SHA256:fyLgNYV+KsCHC/+oAwDLtSgv7PxOz66zG3gSSl7FcfM=')
#                self.assertEqual(l.get_option(ldap0.OPT_X_TLS_PEERKEY_HASH), b'SHA256:o0CxaD0QWxZ4SLxoBan4T+DdDnsVQiWppNsOhakytnQ=')
            l.set_tls_options(cacert_filename='tests/tls/ca.crt')
            l.start_tls_s()
            self.assertEqual(l.whoami_s(), '')
            if hasattr(ldap0, 'OPT_X_TLS_VERSION'):
                self.assertIn(l.get_option(ldap0.OPT_X_TLS_VERSION)[0:5], (b'TLSv1', b'TLS1.'))
            if hasattr(ldap0, 'OPT_X_TLS_CIPHER'):
                self.assertIn(l.get_option(ldap0.OPT_X_TLS_CIPHER)[0:4], (b'TLS_', b'AES-'))
            if hasattr(ldap0, 'OPT_X_TLS_PEERCERT'):
                peer_cert = l.get_option(ldap0.OPT_X_TLS_PEERCERT)
                self.assertIsInstance(peer_cert, bytes)
                self.assertTrue(len(peer_cert) > 1000)
        with self.ldap_object_class(self.server.ldap_uri) as l:
            l.set_tls_options(
                cacert_filename='tests/tls/ca.crt',
                client_cert_filename='tests/tls/test-client.crt',
                client_key_filename='tests/tls/test-client.key',
            )
            l.start_tls_s()
            l.sasl_non_interactive_bind_s('EXTERNAL')
            self.assertEqual(l.whoami_s(), 'dn:cn=test-client,ou=example ou,o=example company')


class Test01_ReconnectLDAPObject(Test00_LDAPObject):
    """
    test ReconnectLDAPObject by restarting slapd
    """

    ldap_object_class = ReconnectLDAPObject

    def test101_reconnect_sasl_external(self):
        with self.ldap_object_class(self.server.ldapi_uri) as l:
            l.sasl_non_interactive_bind_s('EXTERNAL')
            authz_id = l.whoami_s()
            self.assertEqual(authz_id, 'dn:'+self.server.root_dn.lower())
            self.server.restart()
            self.assertEqual(l.whoami_s(), authz_id)

    def test102_reconnect_simple_bind(self):
        with self.ldap_object_class(self.server.ldapi_uri) as l:
            bind_dn = 'cn=user1,'+self.server.suffix
            l.simple_bind_s(bind_dn, b'user1_pw')
            self.assertEqual(l.whoami_s(), 'dn:'+bind_dn)
            self.server.restart()
            self.assertEqual(l.whoami_s(), 'dn:'+bind_dn)

    def test103_reconnect_get_state(self):
        bind_dn = 'cn=user1,'+self.server.suffix
        expected_state = {
            '_last_bind': (
                'simple_bind_s',
                (bind_dn, b'user1_pw'),
                {}
            ),
            '_options': [
                # default options set in LDAPObject.__init__()
                (_libldap0.OPT_PROTOCOL_VERSION, _libldap0.VERSION3),
                (_libldap0.OPT_RESTART, False),
                (_libldap0.OPT_DEREF, False),
                (_libldap0.OPT_REFERRALS, False),
                (_libldap0.OPT_X_TLS_REQUIRE_CERT, _libldap0.OPT_X_TLS_HARD),
            ],
            '_reconnects_done': 0,
            '_retry_delay': 60.0,
            '_retry_max': 1,
            '_start_tls': 0,
            'uri': self.server.ldapi_uri,
            '_cache_ttl': 0.0,
            'timeout': -1,
            '_trace_level': 0,
            'res_call_args': False,
            'encoding': 'utf-8',
        }
        with self.ldap_object_class(self.server.ldapi_uri) as l1:
            l1.simple_bind_s(bind_dn, b'user1_pw')
            self.assertEqual(l1.whoami_s(), 'dn:'+bind_dn)
            self.assertEqual(l1.get_whoami_dn(), bind_dn)
            self.assertEqual(l1.__getstate__(), expected_state)

    def test104_reconnect_restore(self):
        with self.ldap_object_class(self.server.ldapi_uri) as l1:
            l1 = self.ldap_object_class(self.server.ldapi_uri)
            bind_dn = 'cn=user1,'+self.server.suffix
            l1.simple_bind_s(bind_dn, 'user1_pw')
            self.assertEqual(l1.whoami_s(), 'dn:'+bind_dn)
            l1_state = pickle.dumps(l1)
        with pickle.loads(l1_state) as l2:
            self.assertEqual(l2.whoami_s(), 'dn:'+bind_dn)

    def test105_null_reconnect(self):
        with self.ldap_object_class(self.server.ldapi_uri, retry_max=0) as l:
            l.sasl_non_interactive_bind_s('EXTERNAL')
            authz_id = l.whoami_s()
            self.assertEqual(authz_id, 'dn:'+self.server.root_dn.lower())
            self.server.restart()
            with self.assertRaises(_libldap0.SERVER_DOWN):
                self.assertEqual(l.whoami_s(), authz_id)

    def test106_reconnect_reset_last_bind(self):
        with self.ldap_object_class(self.server.ldapi_uri, retry_max=1) as l:
            l.sasl_non_interactive_bind_s('EXTERNAL')
            self.assertEqual(l.whoami_s(), 'dn:'+self.server.root_dn.lower())
            # now force reconnect with anon bind
            l.reconnect(l.uri, reset_last_bind=True)
            self.assertEqual(l.whoami_s(), '')

    def test107_reconnect_failure(self):
        with self.ldap_object_class(self.server.ldapi_uri, retry_max=2, retry_delay=0.1) as l:
            l.sasl_non_interactive_bind_s('EXTERNAL')
            authz_id = l.whoami_s()
            self.assertEqual(authz_id, 'dn:'+self.server.root_dn.lower())
            self.server.stop()
            with self.assertRaises(ldap0.SERVER_DOWN):
                l.whoami_s()
            self.server.start()
            l.reconnect(l.uri)
            l.sasl_non_interactive_bind_s('EXTERNAL')
            self.assertEqual(l.whoami_s(), authz_id)


class TestSchemaUrlFetchLDAP(SlapdTestCase):

    def test_schema_urlfetch(self):
        """
        test ldap0.schema.util.urlfetch() which requires a running slapd instance
        """
        schema_dn, sub_schema = ldap0.schema.util.urlfetch(self.server.ldapi_uri, check_uniqueness=True)
        self.assertEqual(schema_dn, 'cn=Subschema')
        oid = sub_schema.name2oid[ldap0.schema.models.AttributeType]['cn']
        self.assertEqual(oid, '2.5.4.3')
        self.assertEqual(oid, sub_schema.name2oid[ldap0.schema.models.AttributeType]['commonName'])


if __name__ == '__main__':
    unittest.main()



# -*- coding: ascii -*-
"""
ldap0.sasl - support for SASL mechanism
"""

import logging
from typing import Dict


__all__ = [
    'SaslAuth',
    'SaslNoninteractiveAuth',
    'SaslPasswordAuth',
]

# These are the SASL callback id's , as defined in sasl.h
CB_USER = 0x4001
CB_AUTHNAME = 0x4002
CB_LANGUAGE = 0x4003
CB_PASS = 0x4004
CB_ECHOPROMPT = 0x4005
CB_NOECHOPROMPT = 0x4006
CB_GETREALM = 0x4008

# SASL mechs known to only require username and password
SASL_PASSWORD_MECHS = {
    b'CRAM-MD5',
    b'DIGEST-MD5',
    b'LOGIN',
    b'NTLM',
    b'OTP',
    b'PLAIN',
    b'SCRAM-SHA-1',
}

# SASL mechs known to not require application-provided authc information
SASL_NONINTERACTIVE_MECHS = {
    b'EXTERNAL',
    b'GSSAPI',
}


class SaslAuth:
    """
    This class handles SASL interactions for authentication.
    """
    encoding = 'utf-8'

    def __init__(self, cb_value_dict: Dict[int, bytes], trace_level: int = 0):
        self.cb_value_dict = {}
        for key, val in (cb_value_dict or {}).items():
            assert isinstance(val, bytes), TypeError(
                'Expected value for SASL param %d to be bytes, got %r' % (key, val)
            )
            self.cb_value_dict[key] = val
        self._trace_level = trace_level

    def callback(self, cb_id: int, challenge, prompt, defresult) -> bytes:
        """
        callback method invoked by SASL lib
        """
        # The following print command might be useful for debugging
        # new sasl mechanisms. So it is left here
        cb_result = self.cb_value_dict.get(cb_id, defresult) or b''
        if __debug__ and self._trace_level >= 1:
            logging.debug(
                'cb_id=%d, challenge=%r, prompt=%r, defresult=%r -> %r',
                cb_id, challenge, prompt, defresult, cb_result,
            )
        return cb_result


class SaslPasswordAuth(SaslAuth):
    """
    This class handles password-based SASL authentication mechs
    like PLAIN, DIGEST-MD5 or CRAM-MD5 etc.
    """

    def __init__(
            self,
            authc_id: str,
            password: bytes,
            authz_id: str = '',
            trace_level: int = 0,
        ):
        SaslAuth.__init__(
            self,
            {
                CB_AUTHNAME: authc_id.encode(self.encoding),
                CB_PASS: password,
                CB_USER: authz_id.encode(self.encoding),
            },
            trace_level=trace_level,
        )


class SaslNoninteractiveAuth(SaslAuth):
    """
    This class handles non-interactive SASL authentication mechs
    like EXTERNAL or GSSAPI.
    """

    def __init__(
            self,
            authz_id: str = '',
            trace_level: int = 0,
        ):
        SaslAuth.__init__(
            self,
            {
                CB_USER: authz_id.encode(self.encoding),
            },
            trace_level=trace_level,
        )



# -*- coding: ascii -*-
"""
ldap0.controls.vlv - classes for Virtual List View
(see draft-ietf-ldapext-ldapv3-vlv)
"""

from pyasn1.type import univ, namedtype, tag, namedval
from pyasn1.codec.ber import encoder, decoder

from . import RequestControl, ResponseControl


__all__ = [
    'VLVRequestControl',
    'VLVResponseControl',
]


class ByOffsetType(univ.Sequence):
    tagSet = univ.Sequence.tagSet.tagImplicitly(
        tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0)
    )
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('offset', univ.Integer()),
        namedtype.NamedType('contentCount', univ.Integer())
    )


class TargetType(univ.Choice):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('byOffset', ByOffsetType()),
        namedtype.NamedType(
            'greaterThanOrEqual',
            univ.OctetString().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1)
            )
        )
    )


class VirtualListViewRequestType(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('beforeCount', univ.Integer()),
        namedtype.NamedType('afterCount', univ.Integer()),
        namedtype.NamedType('target', TargetType()),
        namedtype.OptionalNamedType('contextID', univ.OctetString())
    )


class VLVRequestControl(RequestControl):
    controlType: str = '2.16.840.1.113730.3.4.9'

    def __init__(
            self,
            criticality: bool = False,
            before_count=0,
            after_count=0,
            offset=None,
            content_count=None,
            greater_than_or_equal=None,
            context_id=None,
        ):
        RequestControl.__init__(self, self.controlType, criticality)
        assert (offset is not None and content_count is not None) or \
               greater_than_or_equal, \
            ValueError(
                'offset and content_count must be set together '
                'or greater_than_or_equal must be used'
            )
        self.before_count = before_count
        self.after_count = after_count
        self.offset = offset
        self.content_count = content_count
        self.greater_than_or_equal = greater_than_or_equal
        self.context_id = context_id

    def encode(self) -> bytes:
        vlv = VirtualListViewRequestType()
        vlv.setComponentByName('beforeCount', self.before_count)
        vlv.setComponentByName('afterCount', self.after_count)
        if self.offset is not None and self.content_count is not None:
            by_offset = ByOffsetType()
            by_offset.setComponentByName('offset', self.offset)
            by_offset.setComponentByName('contentCount', self.content_count)
            target = TargetType()
            target.setComponentByName('byOffset', by_offset)
        elif self.greater_than_or_equal:
            target = TargetType()
            target.setComponentByName('greaterThanOrEqual', self.greater_than_or_equal)
        else:
            raise NotImplementedError
        vlv.setComponentByName('target', target)
        return encoder.encode(vlv)


class VirtualListViewResultType(univ.Enumerated):
    namedValues = namedval.NamedValues(
        ('success', 0),
        ('operationsError', 1),
        ('protocolError', 3),
        ('unwillingToPerform', 53),
        ('insufficientAccessRights', 50),
        ('adminLimitExceeded', 11),
        ('innapropriateMatching', 18),
        ('sortControlMissing', 60),
        ('offsetRangeError', 61),
        ('other', 80),
    )


class VirtualListViewResponseType(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('targetPosition', univ.Integer()),
        namedtype.NamedType('contentCount', univ.Integer()),
        namedtype.NamedType('virtualListViewResult', VirtualListViewResultType()),
        namedtype.OptionalNamedType('contextID', univ.OctetString()),
    )


class VLVResponseControl(ResponseControl):
    controlType: str = '2.16.840.1.113730.3.4.10'

    def __init__(self, criticality=False):
        ResponseControl.__init__(self, self.controlType, criticality)

    def decode(self, encodedControlValue: bytes):
        drv, rest = decoder.decode(
            encodedControlValue,
            asn1Spec=VirtualListViewResponseType(),
        )
        assert not rest, 'all data could not be decoded'
        self.targetPosition = int(drv.getComponentByName('targetPosition'))
        self.contentCount = int(drv.getComponentByName('contentCount'))
        virtual_list_view_result = drv.getComponentByName('virtualListViewResult')
        self.virtualListViewResult = int(virtual_list_view_result)
        context_id = drv.getComponentByName('contextID')
        if context_id.hasValue():
            self.contextID = bytes(context_id)
        else:
            self.contextID = None



# -*- coding: ascii -*-
"""
ldap0.modlist - create add/modify modlist's
"""

from _libldap0 import MOD_DELETE, MOD_ADD


def add_modlist(entry, ignore_attr_types=None):
    """Build modify list for call of method LDAPObject.add()"""
    ignore_attr_types = set(map(str.lower, ignore_attr_types or []))
    modlist = []
    for attrtype in entry:
        if attrtype.lower() in ignore_attr_types:
            # This attribute type is ignored
            continue
        # Eliminate empty attr value strings in list
        attrvaluelist = [
            val
            for val in entry[attrtype]
            if val is not None
        ]
        if attrvaluelist:
            modlist.append((attrtype.encode('ascii'), entry[attrtype]))
    return modlist # add_modlist()


def modify_modlist(
        old_entry,
        new_entry,
        ignore_attr_types=None,
        ignore_oldexistent=0,
        case_ignore_attr_types=None
    ):
    """
    Build differential modify list for calling LDAPObject.modify()/modify_s()

    old_entry
        Dictionary holding the old entry
    new_entry
        Dictionary holding what the new entry should be
    ignore_attr_types
        List of attribute type names to be ignored completely
    ignore_oldexistent
        If non-zero attribute type names which are in old_entry
        but are not found in new_entry at all are not deleted.
        This is handy for situations where your application
        sets attribute value to '' for deleting an attribute.
        In most cases leave zero.
    case_ignore_attr_types
        List of attribute type names for which comparison will be made
        case-insensitive
    """
    ignore_attr_types = set(map(str.lower, ignore_attr_types or []))
    case_ignore_attr_types = set(map(str.lower, case_ignore_attr_types or []))
    modlist = []
    attrtype_lower_map = {}
    for attr_type in old_entry:
        attrtype_lower_map[attr_type.lower()] = attr_type
    for attrtype in new_entry:
        attrtype_b = attrtype.encode('ascii')
        attrtype_lower = attrtype.lower()
        if attrtype_lower in ignore_attr_types:
            # This attribute type is ignored
            continue
        # Filter away null-strings
        new_value = [
            val
            for val in new_entry[attrtype]
            if val is not None
        ]
        if attrtype_lower in attrtype_lower_map:
            old_value = [
                val
                for val in old_entry.get(attrtype_lower_map[attrtype_lower], [])
                if val is not None
            ]
            del attrtype_lower_map[attrtype_lower]
        else:
            old_value = []
        if not old_value and new_value:
            # Add a new attribute to entry
            modlist.append((MOD_ADD, attrtype_b, new_value))
        elif old_value and new_value:
            # Replace existing attribute
            replace_attr_value = len(old_value) != len(new_value)
            if not replace_attr_value:
                if attrtype_lower in case_ignore_attr_types:
                    old_value_set = set(map(bytes.lower, old_value))
                    new_value_set = set(map(bytes.lower, new_value))
                else:
                    old_value_set = set(old_value)
                    new_value_set = set(new_value)
                replace_attr_value = new_value_set != old_value_set
            if replace_attr_value:
                modlist.append((MOD_DELETE, attrtype_b, None))
                modlist.append((MOD_ADD, attrtype_b, new_value))
        elif old_value and not new_value:
            # Completely delete an existing attribute
            modlist.append((MOD_DELETE, attrtype_b, None))
    if not ignore_oldexistent:
        # Remove all attributes of old_entry which are not present
        # in new_entry at all
        for atl in attrtype_lower_map:
            if atl in ignore_attr_types:
                # This attribute type is ignored
                continue
            attrtype = attrtype_lower_map[atl]
            modlist.append((MOD_DELETE, attrtype.encode('ascii'), None))
    return modlist # modify_modlist()



# -*- coding: ascii -*-
"""
ldap0.controls.readentry - classes for the Read Entry controls
(see RFC 4527)
"""

from pyasn1.type import namedtype, univ
from pyasn1.codec.ber import encoder, decoder
from pyasn1.error import PyAsn1Error

from pyasn1_modules.rfc2251 import AttributeDescriptionList, SearchResultEntry
from pyasn1_modules.rfc2251 import LDAPDN, PartialAttributeList

from ..base import encode_list
from ..res import SearchResultEntry as LDAPResultSearchResultEntry
from . import LDAPControl


class OpenLDAPITS6899SearchResultEntry(univ.Sequence):
    """
    This is an ASN.1 description of SearchResultEntry not compliant to LDAPv3
    which implements a work-around for OpenLDAP's ITS#6899
    """
    tagSet = univ.Sequence.tagSet # work-around: instead of implicit tagging
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('objectName', LDAPDN()),
        namedtype.NamedType('attributes', PartialAttributeList())
    )


class ReadEntryControl(LDAPControl):
    """
    Base class for read entry control described in RFC 4527

    attrList
        list of attribute type names requested

    Class attributes with values extracted from the response control:

    dn
        string holding the distinguished name of the LDAP entry
    entry
        dictionary holding the LDAP entry
    """

    def __init__(self, criticality: bool = False, attrList=None):
        LDAPControl.__init__(self, criticality=criticality)
        self.attrList = encode_list(attrList or [], encoding='ascii')

    def encode(self) -> bytes:
        attributeSelection = AttributeDescriptionList()
        for ind in range(len(self.attrList)):
            attributeSelection.setComponentByPosition(ind, self.attrList[ind])
        return encoder.encode(attributeSelection)

    def decode(self, encodedControlValue: bytes):
        try:
            decodedEntry, _ = decoder.decode(
                encodedControlValue,
                asn1Spec=SearchResultEntry(),
            )
        except PyAsn1Error:
            decodedEntry, _ = decoder.decode(
                encodedControlValue,
                asn1Spec=OpenLDAPITS6899SearchResultEntry(),
            )
        self.res = LDAPResultSearchResultEntry(
            bytes(decodedEntry[0]),
            {
                bytes(attr[0]): [
                    bytes(attr_value)
                    for attr_value in attr[1]
                ]
                for attr in decodedEntry[1]
            },
        )

class PreReadControl(ReadEntryControl):
    """
    Class for pre-read control described in RFC 4527
    """
    controlType: str = '1.3.6.1.1.13.1'


class PostReadControl(ReadEntryControl):
    """
    Class for post-read control described in RFC 4527
    """
    controlType: str = '1.3.6.1.1.13.2'



# -*- coding: ascii -*-
"""
ldap0.lock -- specific lock class with logging
"""

import logging
import threading


class LDAPLock:
    """
    Mainly a wrapper class to log all locking events.
    Note that this cumbersome approach with _lock attribute was taken
    since threading.Lock is not suitable for sub-classing.
    """
    __slots__ = (
        '_desc',
        '_lock',
        '_trace_level',
    )
    _min_trace_level: int = 3

    def __init__(self, desc: str, trace_level: int = 0):
        """
        desc
            Description shown in debug log messages
        """
        self._desc = desc
        self._lock = threading.Lock()
        self._trace_level = trace_level

    def __repr__(self) -> str:
        return '%s(%r, %r)' % (self.__class__.__name__, self._desc, self._trace_level)

    def __enter__(self):
        return self._lock.__enter__()

    def __exit__(self, *args):
        return self._lock.__exit__(*args)

    def acquire(self):
        """
        acquire lock and log
        """
        if __debug__:
            if self._trace_level >= self._min_trace_level:
                logging.debug(
                    '%s[%x] acquire %s',
                    self.__class__.__name__,
                    id(self),
                    self._desc,
                )
        return self._lock.acquire()

    def release(self):
        """
        release lock and log
        """
        if __debug__:
            if self._trace_level >= self._min_trace_level:
                logging.debug(
                    '%s[%x] release %s',
                    self.__class__.__name__,
                    id(self),
                    self._desc,
                )
        return self._lock.release()



# -*- coding: ascii -*-
"""
ldap0.cache - dict-like cache class
"""

#-----------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------

# from Python's standard lib
import time

#-----------------------------------------------------------------------
# Constants
#-----------------------------------------------------------------------

__all__ = [
    'Cache',
]


class Cache(dict):

    """
    Class implements a cache dictionary with
    timestamps attached to all dict keys
    """
    __slots__ = (
        '_ttl',
        'hit_count',
        'access_count',
        'max_cache_hit_time',
    )

    def __init__(self, ttl=-1.0):
        """
        :ttl: Time-to-live for cached entries in seconds
        """
        dict.__init__({})
        self._ttl = ttl
        self.hit_count = 0
        self.access_count = 0
        self.max_cache_hit_time = 0.0

    @property
    def miss_count(self) -> int:
        return self.access_count - self.hit_count

    @property
    def hit_ratio(self) -> float:
        """
        Returns percentage of cache hit ratio
        """
        try:
            return 100 * float(self.hit_count) / self.access_count
        except ZeroDivisionError:
            return 0.0

    def __getitem__(self, key):
        self.access_count += 1
        current_time = time.time()
        val, not_after = dict.__getitem__(self, key)
        if current_time > not_after:
            dict.__delitem__(self, key)
            raise KeyError('cache entry expired')
        self.hit_count += 1
        self.max_cache_hit_time = max(
            self.max_cache_hit_time,
            current_time - not_after + self._ttl
        )
        return val

    def __setitem__(self, key, val):
        return dict.__setitem__(
            self,
            key,
            (val, time.time()+self._ttl)
        )

    def cache(self, key, val, ttl=None):
        if ttl is None:
            ttl = self._ttl
        return dict.__setitem__(
            self,
            key,
            (val, time.time()+ttl)
        )

    def flush(self):
        self.hit_count = 0
        self.access_count = 0
        self.clear()



# -*- coding: utf-8 -*-
"""
package/install module package ldap0
"""

import sys
import os
import textwrap
from configparser import ConfigParser
from setuptools import setup, Extension, find_packages

PYPI_NAME = 'ldap0'
BASEDIR = os.path.dirname(os.path.realpath(__file__))

sys.path.insert(0, os.path.join(BASEDIR, PYPI_NAME))
import __about__

class OpenLDAP2BuildConfig:
    """
    class describing the features and requirements of OpenLDAP 2.x
    """
    __slots__ = (
        'defines',
        'extra_compile_args',
        'extra_link_args',
        'extra_objects',
        'include_dirs',
        'library_dirs',
        'libs',
    )

    def __init__(self, meta_defines):
        self.library_dirs = []
        self.include_dirs = []
        self.extra_compile_args = []
        self.extra_link_args = []
        self.extra_objects = []
        self.libs = ['ldap', 'lber']
        self.defines = []
        # Read the [_libldap0] section of setup.cfg
        cfg = ConfigParser()
        cfg.read(os.path.join(BASEDIR, 'setup.cfg'))
        _ldap_cfg = dict(cfg.items('_libldap0'))
        for name in self.__slots__:
            env_name = 'LIBLDAP0_' + name.upper()
            if env_name in os.environ:
                setattr(self, name, [ val for val in os.environ[env_name].split(' ') if val ])
            elif name in _ldap_cfg:
                setattr(self, name, [ val for val in _ldap_cfg[name].split(' ') if val ])
        self.defines = [
            (defm, None)
            for defm in set(self.defines)
        ]
        self.defines.extend(meta_defines)
        self.include_dirs.insert(0, '_libldap0')
        if sys.platform.startswith("win"):
            self.library_dirs = []


LDAP_CLASS = OpenLDAP2BuildConfig(
    [
        ('LDAP0_MODULE_VERSION', __about__.__version__),
        ('LDAP0_MODULE_AUTHOR', __about__.__author__),
        ('LDAP0_MODULE_LICENSE', __about__.__license__),
    ],
)


setup(
    name=PYPI_NAME,
    license=__about__.__license__,
    version=__about__.__version__,
    description='Module package for implementing LDAP clients'.format(PYPI_NAME),
    author=__about__.__author__,
    author_email='michael@stroeder.com',
    maintainer=__about__.__author__,
    maintainer_email='michael@stroeder.com',
    url='https://code.stroeder.com/pymod/python-%s' % (PYPI_NAME),
    download_url='https://pypi.python.org/pypi/%s/' % (PYPI_NAME),
    project_urls={
        'Code': 'https://code.stroeder.com/pymod/python-%s' % (PYPI_NAME),
        'Issue tracker': 'https://code.stroeder.com/pymod/python-%s/issues' % (PYPI_NAME),
    },
    keywords=['LDAP', 'OpenLDAP'],
    ext_modules=[
        Extension(
            name='_libldap0',
            sources=[
                '_libldap0/ldapobject.c',
                '_libldap0/ldapcontrol.c',
                '_libldap0/common.c',
                '_libldap0/constants.c',
                '_libldap0/errors.c',
                '_libldap0/functions.c',
                '_libldap0/libldap0module.c',
                '_libldap0/message.c',
                '_libldap0/options.c',
            ],
            depends = [
                '_libldap0/ldapobject.h',
                '_libldap0/common.h',
                '_libldap0/constants.h',
                '_libldap0/functions.h',
                '_libldap0/ldapcontrol.h',
                '_libldap0/message.h',
                '_libldap0/options.h',
            ],
            libraries=LDAP_CLASS.libs,
            include_dirs=LDAP_CLASS.include_dirs,
            library_dirs=LDAP_CLASS.library_dirs,
            extra_compile_args=LDAP_CLASS.extra_compile_args,
            extra_link_args=LDAP_CLASS.extra_link_args,
            extra_objects=LDAP_CLASS.extra_objects,
            runtime_library_dirs=LDAP_CLASS.library_dirs,
            define_macros=LDAP_CLASS.defines,
        ),
    ],
    packages=find_packages(exclude=['tests']),
    package_dir={'': '.'},
    package_data = {
        PYPI_NAME: ['py.typed'],
    },
    test_suite='tests',
    python_requires='>=3.6',
    include_package_data=True,
    install_requires=[
        'setuptools',
        'pyasn1>=0.4.2',
        'pyasn1_modules>=0.2.1',
    ],
    zip_safe=True,
)



# -*- coding: utf-8 -*-
"""
Automatic tests for module ldap0.ldif
"""

# from Python's standard lib
import unittest
import textwrap

from io import BytesIO

from ldap0.ldif import LDIFParser, LDIFWriter, MOD_OP_INTEGER


class TestLDIFParser(unittest.TestCase):
    """
    Various LDIF test cases
    """
    record_type = None

    def _parse_records(
            self,
            ldif_string,
            ignored_attr_types=None,
            max_entries=None,
        ):
        """
        Parse LDIF data in `ldif_string' into list of records
        """
        ldif_parser = LDIFParser.frombuf(
            ldif_string,
            ignored_attr_types=ignored_attr_types,
            max_entries=max_entries,
        )
        parser_method = getattr(
            ldif_parser,
            'list_{0}_records'.format(self.record_type)
        )
        return parser_method()

    def _unparse_records(self, records):
        """
        Returns LDIF string with entry records from list `records'
        """
        ldif_file = BytesIO()
        ldif_writer = LDIFWriter(ldif_file)
        if self.record_type == 'entry':
            for dn, entry in records:
                ldif_writer.unparse(dn, entry)
        elif self.record_type == 'change':
            for dn, modops, controls in records:
                ldif_writer.unparse(dn, modops)
        self.assertEqual(ldif_writer.records_written, len(records))
        return ldif_file.getvalue()

    def check_records(
            self,
            ldif_string,
            records,
            ignored_attr_types=None,
            max_entries=None,
    ):
        """
        Checks whether entry records in `ldif_string' gets correctly parsed
        and matches list of unparsed `records'.
        """
        ldif_string = textwrap.dedent(ldif_string).lstrip()
        parsed_records = self._parse_records(
            ldif_string.encode('utf-8'),
            ignored_attr_types=ignored_attr_types,
            max_entries=max_entries,
        )
        generated_ldif = self._unparse_records(records)
        parsed_records2 = self._parse_records(
            generated_ldif,
            ignored_attr_types=ignored_attr_types,
            max_entries=max_entries,
        )
        self.assertEqual(records, parsed_records)
        self.assertEqual(records, parsed_records2)


class TestEntryRecords(TestLDIFParser):
    """
    Various LDIF test cases
    """
    record_type='entry'

    def test_yield(self):
        ldif_string = \
            """
            # comment #1
             with line-folding
            dn: cn=x1,cn=y1,cn=z1
            b1: value_b1
            c1: value_c1
            a1: value_a1

            # comment #2.1
            # comment #2.2
            dn: cn=x2,cn=y2,cn=z2
            b2: value_b2
            c2: value_c2
            a2: value_a2

            """
        ldif_string = textwrap.dedent(ldif_string).lstrip()
        result = LDIFParser.frombuf(ldif_string.encode('utf-8')).list_entry_records()

    def test_empty(self):
        self.check_records(
            """
            version: 1

            """,
            []
        )

    def test_simple(self):
        self.check_records(
            """
            version: 1

            dn: cn=x,cn=y,cn=z
            attrib: value
            attrib: value2

            """,
            [
                (
                    b'cn=x,cn=y,cn=z',
                    {
                        b'attrib': [b'value', b'value2'],
                    },
                ),
            ]
        )

    def test_simple2(self):
        self.check_records(
            """
            dn:cn=x,cn=y,cn=z
            attrib:value
            attrib:value2

            """,
            [
                (
                    b'cn=x,cn=y,cn=z',
                    {
                        b'attrib': [b'value', b'value2'],
                    },
                ),
            ]
        )

    def test_multiple(self):
        self.check_records(
            """
            dn: cn=x,cn=y,cn=z
            a: v
            attrib: value
            attrib: value2

            dn: cn=a,cn=b,cn=c
            attrib: value2
            attrib: value3
            b: v

            """,
            [
                (
                    b'cn=x,cn=y,cn=z',
                    {
                        b'attrib': [b'value', b'value2'],
                        b'a': [b'v'],
                    },
                ),
                (
                    b'cn=a,cn=b,cn=c',
                    {
                        b'attrib': [b'value2', b'value3'],
                        b'b': [b'v'],
                    },
                ),
            ]
        )

    def test_folded(self):
        self.check_records(
            """
            dn: cn=x,cn=y,cn=z
            attrib: very\x20
             long
              line-folded\x20
             value
            attrib2: %s

            """ % ('asdf.'*20),
            [
                (
                    b'cn=x,cn=y,cn=z',
                    {
                        b'attrib': [b'very long line-folded value'],
                        b'attrib2': [b'asdf.'*20],
                    }
                ),
            ]
        )

    def test_empty_attr_values(self):
        self.check_records(
            """
            dn: cn=x,cn=y,cn=z
            attrib1:
            attrib1: foo
            attrib2:
            attrib2: foo

            """,
            [
                (
                    b'cn=x,cn=y,cn=z',
                    {
                        b'attrib1': [b'', b'foo'],
                        b'attrib2': [b'', b'foo'],
                    },
                ),
            ]
        )

    def test_binary(self):
        self.check_records(
            """
            dn: cn=x,cn=y,cn=z
            attrib:: CQAKOiVA

            """,
            [
                (
                    b'cn=x,cn=y,cn=z',
                    {
                        b'attrib': [b'\t\0\n:%@'],
                    },
                ),
            ]
        )

    def test_binary2(self):
        self.check_records(
            """
            dn: cn=x,cn=y,cn=z
            attrib::CQAKOiVA

            """,
            [
                (
                    b'cn=x,cn=y,cn=z',
                    {b'attrib': [b'\t\0\n:%@']},
                ),
            ]
        )

    def test_big_binary(self):
        self.check_records(
            """
            dn: cn=x,cn=y,cn=z
            attrib:: AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
             AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
             AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
             AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
             AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
             AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
             AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
             AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
             AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
             =

            """,
            [
                (
                    b'cn=x,cn=y,cn=z',
                    {b'attrib': [500*b'\0']},
                ),
            ]
        )

    def test_unicode(self):
        self.check_records(
            """
            dn: cn=Michael Ströder,dc=stroeder,dc=com
            cn: Michael Ströder
            sn: Ströder

            """,
            [
                (
                    b'cn=Michael Str\xc3\xb6der,dc=stroeder,dc=com',
                    {
                        b'cn': [b'Michael Str\xc3\xb6der'],
                        b'sn': [b'Str\xc3\xb6der'],
                    },
                ),
            ]
        )

    def test_sorted(self):
        self.check_records(
            """
            dn: cn=x,cn=y,cn=z
            b: value_b
            c: value_c
            a: value_a

            """,
            [
                (
                    b'cn=x,cn=y,cn=z',
                    {
                        b'a': [b'value_a'],
                        b'b': [b'value_b'],
                        b'c': [b'value_c'],
                    }
                ),
            ]
        )

    def test_ignored_attr_types(self):
        self.check_records(
            """
            dn: cn=x,cn=y,cn=z
            a: value_a
            b: value_b
            c: value_c

            """,
            [
                (
                    b'cn=x,cn=y,cn=z',
                    {
                        b'a': [b'value_a'],
                        b'c': [b'value_c'],
                    }
                ),
            ],
            ignored_attr_types=[b'b'],
        )

    def test_comments(self):
        self.check_records(
            """
            # comment #1
             with line-folding
            dn: cn=x1,cn=y1,cn=z1
            b1: value_b1
            c1: value_c1
            a1: value_a1

            # comment #2.1
            # comment #2.2
            dn: cn=x2,cn=y2,cn=z2
            b2: value_b2
            c2: value_c2
            a2: value_a2

            """,
            [
                (
                    b'cn=x1,cn=y1,cn=z1',
                    {
                        b'a1': [b'value_a1'],
                        b'b1': [b'value_b1'],
                        b'c1': [b'value_c1'],
                    }
                ),
                (
                    b'cn=x2,cn=y2,cn=z2',
                    {
                        b'a2': [b'value_a2'],
                        b'b2': [b'value_b2'],
                        b'c2': [b'value_c2'],
                    }
                ),
            ]
        )

    def test_max_entries(self):
        self.check_records(
            """
            dn: cn=x1,cn=y1,cn=z1
            b1: value_b1
            a1: value_a1

            dn: cn=x2,cn=y2,cn=z2
            b2: value_b2
            a2: value_a2

            dn: cn=x3,cn=y3,cn=z3
            b3: value_b3
            a3: value_a3

            dn: cn=x4,cn=y4,cn=z4
            b2: value_b4
            a2: value_a4

            """,
            [
                (
                    b'cn=x1,cn=y1,cn=z1',
                    {
                        b'a1': [b'value_a1'],
                        b'b1': [b'value_b1'],
                    }
                ),
                (
                    b'cn=x2,cn=y2,cn=z2',
                    {
                        b'a2': [b'value_a2'],
                        b'b2': [b'value_b2'],
                    }
                ),
            ],
            max_entries=2
        )

    def test_missing_trailing_line_separator(self):
        self.check_records(
            """
            dn: cn=x1,cn=y1,cn=z1
            first: value_a1
            middle: value_b1
            last: value_c1

            dn: cn=x2,cn=y2,cn=z2
            first: value_a2
            middle: value_b2
            last: value_c2""",
            [
                (
                    b'cn=x1,cn=y1,cn=z1',
                    {
                        b'first': [b'value_a1'],
                        b'middle': [b'value_b1'],
                        b'last': [b'value_c1'],
                    }
                ),
                (
                    b'cn=x2,cn=y2,cn=z2',
                    {
                        b'first': [b'value_a2'],
                        b'middle': [b'value_b2'],
                        b'last': [b'value_c2'],
                    }
                ),
            ],
        )

    def test_weird_empty_lines(self):
        self.check_records(
            """

            # comment before version

            version: 1


            dn: cn=x1,cn=y1,cn=z1
            first: value_a1
            middle: value_b1
            last: value_c1


            dn: cn=x2,cn=y2,cn=z2
            first: value_a2
            middle: value_b2
            last: value_c2""",
            [
                (
                    b'cn=x1,cn=y1,cn=z1',
                    {
                        b'first': [b'value_a1'],
                        b'middle': [b'value_b1'],
                        b'last': [b'value_c1'],
                    }
                ),
                (
                    b'cn=x2,cn=y2,cn=z2',
                    {
                        b'first': [b'value_a2'],
                        b'middle': [b'value_b2'],
                        b'last': [b'value_c2'],
                    }
                ),
            ],
        )

    def test_multiple_empty_lines(self):
        """
        test malformed LDIF with multiple empty lines
        """
        self.check_records(
            """
            # normal
            dn: uid=one,dc=tld
            uid: one



            # after extra empty line
            dn: uid=two,dc=tld
            uid: two

            """,
            [
                (
                    b'uid=one,dc=tld',
                    {b'uid': [b'one']}
                ),
                (
                    b'uid=two,dc=tld',
                    {b'uid': [b'two']}
                ),
            ],
        )


class TestChangeRecords(TestLDIFParser):
    """
    Various LDIF test cases
    """
    record_type='change'

    def test_empty(self):
        self.check_records(
            """
            version: 1
            """,
            [],
        )

    def test_simple(self):
        self.check_records(
            """
            version: 1

            dn: cn=x,cn=y,cn=z
            changetype: modify
            replace: attrib
            attrib: value
            attrib: value2
            -
            add: attrib2
            attrib2: value
            attrib2: value2
            -
            delete: attrib3
            attrib3: value
            -
            delete: attrib4
            -

            """,
            [
                (
                    b'cn=x,cn=y,cn=z',
                    [
                        (MOD_OP_INTEGER[b'replace'], b'attrib', [b'value', b'value2']),
                        (MOD_OP_INTEGER[b'add'], b'attrib2', [b'value', b'value2']),
                        (MOD_OP_INTEGER[b'delete'], b'attrib3', [b'value']),
                        (MOD_OP_INTEGER[b'delete'], b'attrib4', None),
                    ],
                    [],
                ),
            ],
        )

    def test_weird_empty_lines(self):
        self.check_records(
            """

            # comment before version

            version: 1


            dn: cn=x,cn=y,cn=z
            changetype: modify
            replace: attrib
            attrib: value
            attrib: value2
            -
            add: attrib2
            attrib2: value
            attrib2: value2
            -
            delete: attrib3
            attrib3: value
            -
            delete: attrib4
            -


            dn: cn=foo,cn=bar
            changetype: modify
            replace: attrib
            attrib: value
            attrib: value2
            -
            add: attrib2
            attrib2: value
            attrib2: value2
            -
            delete: attrib3
            attrib3: value
            -
            delete: attrib4""",
            [
                (
                    b'cn=x,cn=y,cn=z',
                    [
                        (MOD_OP_INTEGER[b'replace'], b'attrib', [b'value', b'value2']),
                        (MOD_OP_INTEGER[b'add'], b'attrib2', [b'value', b'value2']),
                        (MOD_OP_INTEGER[b'delete'], b'attrib3', [b'value']),
                        (MOD_OP_INTEGER[b'delete'], b'attrib4', None),
                    ],
                    [],
                ),
                (
                    b'cn=foo,cn=bar',
                    [
                        (MOD_OP_INTEGER[b'replace'], b'attrib', [b'value', b'value2']),
                        (MOD_OP_INTEGER[b'add'], b'attrib2', [b'value', b'value2']),
                        (MOD_OP_INTEGER[b'delete'], b'attrib3', [b'value']),
                        (MOD_OP_INTEGER[b'delete'], b'attrib4', None),
                    ],
                    [],
                ),
            ],
        )

    def test_missing_trailing_dash_separator(self):
        self.check_records(
            """
            version: 1

            dn: cn=x,cn=y,cn=z
            changetype: modify
            replace: attrib
            attrib: value
            attrib: value2
            -
            add: attrib2
            attrib2: value
            attrib2: value2

            """,
            [
                (
                    b'cn=x,cn=y,cn=z',
                    [
                        (MOD_OP_INTEGER[b'replace'], b'attrib', [b'value', b'value2']),
                        (MOD_OP_INTEGER[b'add'], b'attrib2', [b'value', b'value2']),
                    ],
                    [],
                ),
            ],
        )

    def test_bad_change_records(self):
        for bad_ldif_string in (
            """
            changetype: modify
            replace: attrib
            attrib: value
            attrib: value2

            """,
            """
            dn: cn=foobar
            changetype: modify
            foo: attrib
            attrib: value
            attrib: value2

            """,
            """
            dn: cn=foobar
            changetype: modify
            add: attrib
            attrib: value
            attrib: value2
            \t-

            """,
        ):
            ldif_string = textwrap.dedent(bad_ldif_string).lstrip() + '\n'
            with self.assertRaises(ValueError):
                self._parse_records(ldif_string.encode('utf-8'))

    def test_mod_increment(self):
        self.check_records(
            """
            version: 1

            dn: cn=x,cn=y,cn=z
            changetype: modify
            increment: gidNumber
            gidNumber: 1
            -

            """,
            [
                (
                    b'cn=x,cn=y,cn=z',
                    [
                        (MOD_OP_INTEGER[b'increment'], b'gidNumber', [b'1']),
                    ],
                    [],
                ),
            ],
        )


if __name__ == '__main__':
    unittest.main()



# -*- coding: ascii -*-
"""
ldap0.__about__ - Meta information
"""

import collections

VersionInfo = collections.namedtuple('VersionInfo', ('major', 'minor', 'micro'))
__version_info__ = VersionInfo(
    major=1,
    minor=4,
    micro=10,
)
__version__ = '.'.join(str(val) for val in __version_info__)
__author__ = 'Michael Stroeder'
__mail__ = 'michael@stroeder.com'
__copyright__ = '(C) 2017-2022 by Michael Str\xF6der <michael@stroeder.com>'
__license__ = 'Apache-2.0'

__all__ = [
    '__version_info__',
    '__version__',
    '__author__',
    '__mail__',
    '__license__',
    '__copyright__',
]



# -*- coding: ascii -*-
"""
ldap0.schema.models - support for subSchemaSubEntry information
"""

from collections.abc import MutableSet
from collections import UserDict, OrderedDict
from typing import Any, Mapping, Sequence

from ..cidict import CIDict
from .parse import split_tokens, extract_tokens


__all__ = [
    'AttributeType',
    'DITContentRule',
    'DITStructureRule',
    'Entry',
    'LDAPSyntax',
    'MatchingRule',
    'MatchingRuleUse',
    'NameForm',
    'ObjectClass',
    'SchemaElement',
    'SchemaElementOIDSet',
    'NOT_HUMAN_READABLE_LDAP_SYNTAXES',
    'OBJECTCLASS_KIND_STR',
]

#TokenDefaults = Mapping[str, Union[int, str, Tuple[Union[None, str]]]]
TokenDefaults = Mapping[str, Any]

NOT_HUMAN_READABLE_LDAP_SYNTAXES = {
    '1.3.6.1.4.1.1466.115.121.1.4',  # Audio
    '1.3.6.1.4.1.1466.115.121.1.5',  # Binary
    '1.3.6.1.4.1.1466.115.121.1.8',  # Certificate
    '1.3.6.1.4.1.1466.115.121.1.9',  # Certificate List
    '1.3.6.1.4.1.1466.115.121.1.10', # Certificate Pair
    '1.3.6.1.4.1.1466.115.121.1.23', # G3 FAX
    '1.3.6.1.4.1.1466.115.121.1.28', # JPEG
    '1.3.6.1.4.1.1466.115.121.1.40', # Octet String
    '1.3.6.1.4.1.1466.115.121.1.49', # Supported Algorithm
}

ATTRIBUTE_USAGE = CIDict({
    'userApplication': 0, # work-around for non-compliant schema
    'userApplications': 0,
    'directoryOperation': 1,
    'distributedOperation': 2,
    'dSAOperation': 3,
})

OBJECTCLASS_KIND_STR = {
    0: 'STRUCTURAL',
    1: 'ABSTRACT',
    2: 'AUXILIARY'
}


class SchemaElement:
    """
    Base class for all schema element classes. Not used directly!

    Arguments:

    schema_element_str
      String which contains the schema element description to be parsed.

    Class attributes:

    schema_attribute
      LDAP attribute type containing a certain schema element description
    token_defaults
      Dictionary internally used by the schema element parser
      containing the defaults for certain schema description key-words
    """
    __slots__ = (
        'oid',
        'desc',
        'obsolete',
    )
    token_defaults: TokenDefaults = {
        'DESC':(None,),
    }
    oid: str
    desc: str

    def __init__(self, schema_element_str=None):
        if schema_element_str:
            tkl = split_tokens(schema_element_str)
            self.set_id(tkl[1])
            opts = extract_tokens(tkl, self.token_defaults)
            self._set_attrs(opts)

    def _set_attrs(self, opts):
        self.desc = opts['DESC'][0]

    def set_id(self, element_id):
        self.oid = element_id

    def get_id(self):
        return self.oid

    @staticmethod
    def key_attr(key, value, quoted=0):
        assert value is None or isinstance(value, str), \
            TypeError("value has to be of str, was %r" % value)
        if not value:
            return ""
        if quoted:
            return " %s '%s'" % (key, value.replace("'", "\\'"))
        return " %s %s" % (key, value)

    @staticmethod
    def key_list(key, values, sep=' ', quoted=0):
        assert isinstance(values, tuple), TypeError("values has to be a tuple, was %r" % values)
        if not values:
            return ''
        if quoted:
            quoted_values = ["'%s'" % value.replace("'", "\\'") for value in values]
        else:
            quoted_values = values
        if len(values) == 1:
            return ' %s %s' % (key, quoted_values[0])
        return ' %s ( %s )' % (key, sep.join(quoted_values))

    def __str__(self):
        result = [str(self.oid)]
        result.append(self.key_attr('DESC', self.desc, quoted=1))
        return '( %s )' % ''.join(result)


class ObjectClass(SchemaElement):
    """
    Arguments:

    schema_element_str
      String containing an ObjectClassDescription

    Class attributes:

    oid
      OID assigned to the object class
    names
      This list of strings contains all NAMEs of the object class
    desc
      This string contains description text (DESC) of the object class
    obsolete
      Integer flag (0 or 1) indicating whether the object class is marked
      as OBSOLETE in the schema
    must
      This list of strings contains NAMEs or OIDs of all attributes
      an entry of the object class must have
    may
      This list of strings contains NAMEs or OIDs of additional attributes
      an entry of the object class may have
    kind
      Kind of an object class:
      0 = STRUCTURAL,
      1 = ABSTRACT,
      2 = AUXILIARY
    sup
      This list of strings contains NAMEs or OIDs of object classes
      this object class is derived from
    """
    __slots__ = (
        'names',
        'obsolete',
        'sup',
        'must',
        'may',
        'kind',
    )
    schema_attribute = 'objectClasses'
    token_defaults: TokenDefaults = {
        'NAME': (()),
        'DESC': (None,),
        'OBSOLETE': None,
        'SUP': (()),
        'STRUCTURAL': None,
        'AUXILIARY': None,
        'ABSTRACT': None,
        'MUST': (()),
        'MAY': ()
    }
    names: Sequence[str]

    def _set_attrs(self, opts):
        self.obsolete = opts['OBSOLETE'] is not None
        self.names = opts['NAME']
        self.desc = opts['DESC'][0]
        self.must = opts['MUST']
        self.may = opts['MAY']
        # Default is STRUCTURAL, see RFC 4512
        self.kind = 0
        if opts['ABSTRACT'] is not None:
            self.kind = 1
        elif opts['AUXILIARY'] is not None:
            self.kind = 2
        if self.kind == 0 and not opts['SUP'] and self.oid != '2.5.6.0':
            # STRUCTURAL object classes are sub-classes of 'top' by default
            self.sup = ('top',)
        else:
            self.sup = opts['SUP']

    def __str__(self):
        result = [str(self.oid)]
        result.append(self.key_list('NAME', self.names, quoted=1))
        result.append(self.key_attr('DESC', self.desc, quoted=1))
        result.append(self.key_list('SUP', self.sup, sep=' $ '))
        result.append({0:'', 1:' OBSOLETE'}[self.obsolete])
        result.append(' %s' % OBJECTCLASS_KIND_STR[self.kind])
        result.append(self.key_list('MUST', self.must, sep=' $ '))
        result.append(self.key_list('MAY', self.may, sep=' $ '))
        return '( %s )' % ''.join(result)


class AttributeType(SchemaElement):
    """
    Arguments:

    schema_element_str
      String containing an AttributeTypeDescription

    Class attributes:

    oid
      OID assigned to the attribute type
    names
      This list of strings contains all NAMEs of the attribute type
    desc
      This string contains description text (DESC) of the attribute type
    obsolete
      Integer flag (0 or 1) indicating whether the attribute type is marked
      as OBSOLETE in the schema
    single_value
      Integer flag (0 or 1) indicating whether the attribute must
      have only one value
    syntax
      String contains OID of the LDAP syntax assigned to the attribute type
    no_user_mod
      Integer flag (0 or 1) indicating whether the attribute is modifiable
      by a client application
    equality
      String contains NAME or OID of the matching rule used for
      checking whether attribute values are equal
    substr
      String contains NAME or OID of the matching rule used for
      checking whether an attribute value contains another value
    ordering
      String contains NAME or OID of the matching rule used for
      checking whether attribute values are lesser-equal than
    usage
      USAGE of an attribute type:
      0 = userApplications
      1 = directoryOperation,
      2 = distributedOperation,
      3 = dSAOperation
    sup
      This list of strings contains NAMEs or OIDs of attribute types
      this attribute type is derived from
    """
    __slots__ = (
        'names',
        'obsolete',
        'sup',
        'equality',
        'ordering',
        'syntax',
        'syntax_len',
        'single_value',
        'collective',
        'substr',
        'x_origin',
        'x_ordered',
        'no_user_mod',
        'usage',
    )
    schema_attribute = 'attributeTypes'
    token_defaults: TokenDefaults = {
        'NAME': (()),
        'DESC': (None,),
        'OBSOLETE': None,
        'SUP': (()),
        'EQUALITY': (None,),
        'ORDERING': (None,),
        'SUBSTR': (None,),
        'SYNTAX': (None,),
        'SINGLE-VALUE': None,
        'COLLECTIVE': None,
        'NO-USER-MODIFICATION': None,
        'USAGE': ('userApplications',),
        'X-ORIGIN': (None,),
        'X-ORDERED': (None,),
    }

    def _set_attrs(self, opts):
        self.names = opts['NAME']
        self.desc = opts['DESC'][0]
        self.obsolete = opts['OBSOLETE'] is not None
        self.sup = opts['SUP']
        self.equality = opts['EQUALITY'][0]
        self.ordering = opts['ORDERING'][0]
        self.substr = opts['SUBSTR'][0]
        self.x_origin = opts['X-ORIGIN'][0]
        self.x_ordered = opts['X-ORDERED'][0]
        try:
            syntax = opts['SYNTAX'][0]
        except IndexError:
            self.syntax = None
            self.syntax_len = None
        else:
            if syntax is None:
                self.syntax = None
                self.syntax_len = None
            else:
                try:
                    self.syntax, syntax_len = opts['SYNTAX'][0].split("{")
                except ValueError:
                    self.syntax = opts['SYNTAX'][0]
                    self.syntax_len = None
                else:
                    self.syntax_len = int(syntax_len[:-1])
        self.single_value = opts['SINGLE-VALUE'] is not None
        self.collective = opts['COLLECTIVE'] is not None
        self.no_user_mod = opts['NO-USER-MODIFICATION'] is not None
        self.usage = ATTRIBUTE_USAGE.get(opts['USAGE'][0], 0)

    def __str__(self):
        result = [str(self.oid)]
        result.append(self.key_list('NAME', self.names, quoted=1))
        result.append(self.key_attr('DESC', self.desc, quoted=1))
        result.append(self.key_list('SUP', self.sup, sep=' $ '))
        result.append({0:'', 1:' OBSOLETE'}[self.obsolete])
        result.append(self.key_attr('EQUALITY', self.equality))
        result.append(self.key_attr('ORDERING', self.ordering))
        result.append(self.key_attr('SUBSTR', self.substr))
        result.append(self.key_attr('SYNTAX', self.syntax))
        if self.syntax_len is not None:
            result.append(('{%d}' % (self.syntax_len))*(self.syntax_len > 0))
        result.append({0:'', 1:' SINGLE-VALUE'}[self.single_value])
        result.append({0:'', 1:' COLLECTIVE'}[self.collective])
        result.append({0:'', 1:' NO-USER-MODIFICATION'}[self.no_user_mod])
        result.append(
            {
                0: "",
                1: " USAGE directoryOperation",
                2: " USAGE distributedOperation",
                3: " USAGE dSAOperation",
            }[self.usage]
        )
        result.append(self.key_attr('X-ORIGIN', self.x_origin, quoted=1))
        result.append(self.key_attr('X-ORDERED', self.x_ordered, quoted=1))
        return '( %s )' % ''.join(result)


class LDAPSyntax(SchemaElement):
    """
    SyntaxDescription

    oid
      OID assigned to the LDAP syntax
    desc
      This string contains description text (DESC) of the LDAP syntax
    not_human_readable
      Integer flag (0 or 1) indicating whether the attribute type is marked
      as not human-readable (X-NOT-HUMAN-READABLE)
    """
    __slots__ = (
        'x_subst',
        'not_human_readable',
        'x_binary_transfer_required',
    )
    schema_attribute = 'ldapSyntaxes'
    token_defaults: TokenDefaults = {
        'DESC': (None,),
        'X-NOT-HUMAN-READABLE': (None,),
        'X-BINARY-TRANSFER-REQUIRED': (None,),
        'X-SUBST': (None,),
    }

    def _set_attrs(self, opts):
        self.desc = opts['DESC'][0]
        self.x_subst = opts['X-SUBST'][0]
        self.not_human_readable = \
            self.oid in NOT_HUMAN_READABLE_LDAP_SYNTAXES or \
            opts['X-NOT-HUMAN-READABLE'][0] == 'TRUE'
        self.x_binary_transfer_required = opts['X-BINARY-TRANSFER-REQUIRED'][0] == 'TRUE'

    def __str__(self):
        result = [str(self.oid)]
        result.append(self.key_attr('DESC', self.desc, quoted=1))
        result.append(self.key_attr('X-SUBST', self.x_subst, quoted=1))
        result.append(
            {0:'', 1:" X-NOT-HUMAN-READABLE 'TRUE'"}[self.not_human_readable]
        )
        result.append(
            {0:'', 1:" X-BINARY-TRANSFER-REQUIRED 'TRUE'"}[self.x_binary_transfer_required]
        )
        return '( %s )' % ''.join(result)


class MatchingRule(SchemaElement):
    """
    Arguments:

    schema_element_str
      String containing an MatchingRuleDescription

    Class attributes:

    oid
      OID assigned to the matching rule
    names
      This list of strings contains all NAMEs of the matching rule
    desc
      This string contains description text (DESC) of the matching rule
    obsolete
      Integer flag (0 or 1) indicating whether the matching rule is marked
      as OBSOLETE in the schema
    syntax
      String contains OID of the LDAP syntax this matching rule is usable with
    """
    __slots__ = (
        'names',
        'syntax',
    )
    schema_attribute = 'matchingRules'
    token_defaults: TokenDefaults = {
        'NAME': (()),
        'DESC': (None,),
        'OBSOLETE': None,
        'SYNTAX': (None,),
    }

    def _set_attrs(self, opts):
        self.names = opts['NAME']
        self.desc = opts['DESC'][0]
        self.obsolete = opts['OBSOLETE'] is not None
        self.syntax = opts['SYNTAX'][0]

    def __str__(self):
        result = [str(self.oid)]
        result.append(self.key_list('NAME', self.names, quoted=1))
        result.append(self.key_attr('DESC', self.desc, quoted=1))
        result.append({0:'', 1:' OBSOLETE'}[self.obsolete])
        result.append(self.key_attr('SYNTAX', self.syntax))
        return '( %s )' % ''.join(result)


class MatchingRuleUse(SchemaElement):
    """
    Arguments:

    schema_element_str
      String containing an MatchingRuleUseDescription

    Class attributes:

    oid
      OID of the accompanying matching rule
    names
      This list of strings contains all NAMEs of the matching rule
    desc
      This string contains description text (DESC) of the matching rule
    obsolete
      Integer flag (0 or 1) indicating whether the matching rule is marked
      as OBSOLETE in the schema
    applies
      This list of strings contains NAMEs or OIDs of attribute types
      for which this matching rule is used
    """
    __slots__ = (
        'names',
        'applies',
    )
    schema_attribute = 'matchingRuleUse'
    token_defaults: TokenDefaults = {
        'NAME': (()),
        'DESC': (None,),
        'OBSOLETE': None,
        'APPLIES': (()),
    }

    def _set_attrs(self, opts):
        self.names = opts['NAME']
        self.desc = opts['DESC'][0]
        self.obsolete = opts['OBSOLETE'] is not None
        self.applies = opts['APPLIES']

    def __str__(self):
        result = [str(self.oid)]
        result.append(self.key_list('NAME', self.names, quoted=1))
        result.append(self.key_attr('DESC', self.desc, quoted=1))
        result.append({0:'', 1:' OBSOLETE'}[self.obsolete])
        result.append(self.key_list('APPLIES', self.applies, sep=' $ '))
        return '( %s )' % ''.join(result)


class DITContentRule(SchemaElement):
    """
    Arguments:

    schema_element_str
      String containing an DITContentRuleDescription

    Class attributes:

    oid
      OID of the accompanying structural object class
    names
      This list of strings contains all NAMEs of the DIT content rule
    desc
      This string contains description text (DESC) of the DIT content rule
    obsolete
      Integer flag (0 or 1) indicating whether the DIT content rule is marked
      as OBSOLETE in the schema
    aux
      This list of strings contains NAMEs or OIDs of all auxiliary
      object classes usable in an entry of the object class
    must
      This list of strings contains NAMEs or OIDs of all attributes
      an entry of the object class must have which may extend the
      list of required attributes of the object classes of an entry
    may
      This list of strings contains NAMEs or OIDs of additional attributes
      an entry of the object class may have which may extend the
      list of optional attributes of the object classes of an entry
    nots
      This list of strings contains NAMEs or OIDs of attributes which
      may not be present in an entry of the object class
    """
    __slots__ = (
        'names',
        'aux',
        'must',
        'may',
        'nots',
    )
    schema_attribute = 'dITContentRules'
    token_defaults: TokenDefaults = {
        'NAME': (()),
        'DESC': (None,),
        'OBSOLETE': None,
        'AUX': (()),
        'MUST': (()),
        'MAY': (()),
        'NOT': (()),
    }

    def _set_attrs(self, opts):
        self.names = opts['NAME']
        self.desc = opts['DESC'][0]
        self.obsolete = opts['OBSOLETE'] is not None
        self.aux = opts['AUX']
        self.must = opts['MUST']
        self.may = opts['MAY']
        self.nots = opts['NOT']

    def __str__(self):
        result = [str(self.oid)]
        result.append(self.key_list('NAME', self.names, quoted=1))
        result.append(self.key_attr('DESC', self.desc, quoted=1))
        result.append({0:'', 1:' OBSOLETE'}[self.obsolete])
        result.append(self.key_list('AUX', self.aux, sep=' $ '))
        result.append(self.key_list('MUST', self.must, sep=' $ '))
        result.append(self.key_list('MAY', self.may, sep=' $ '))
        result.append(self.key_list('NOT', self.nots, sep=' $ '))
        return '( %s )' % ''.join(result)


class DITStructureRule(SchemaElement):
    """
    Arguments:

    schema_element_str
      String containing an DITStructureRuleDescription

    Class attributes:

    ruleid
      rule ID of the DIT structure rule (only locally unique)
    names
      This list of strings contains all NAMEs of the DIT structure rule
    desc
      This string contains description text (DESC) of the DIT structure rule
    obsolete
      Integer flag (0 or 1) indicating whether the DIT content rule is marked
      as OBSOLETE in the schema
    form
      List of strings with NAMEs or OIDs of associated name forms
    sup
      List of strings with NAMEs or OIDs of allowed structural object classes
      of superior entries in the DIT
    """
    __slots__ = (
        'ruleid',
        'names',
        'sup',
        'form',
    )
    schema_attribute = 'dITStructureRules'
    token_defaults: TokenDefaults = {
        'NAME': (()),
        'DESC': (None,),
        'OBSOLETE': None,
        'FORM': (None,),
        'SUP': (()),
    }
    ruleid: str

    def set_id(self, element_id):
        self.ruleid = element_id

    def get_id(self):
        return self.ruleid

    def _set_attrs(self, opts):
        self.names = opts['NAME']
        self.desc = opts['DESC'][0]
        self.obsolete = opts['OBSOLETE'] is not None
        self.form = opts['FORM'][0]
        self.sup = opts['SUP']

    def __str__(self):
        result = [str(self.ruleid)]
        result.append(self.key_list('NAME', self.names, quoted=1))
        result.append(self.key_attr('DESC', self.desc, quoted=1))
        result.append({0:'', 1:' OBSOLETE'}[self.obsolete])
        result.append(self.key_attr('FORM', self.form, quoted=0))
        result.append(self.key_list('SUP', self.sup, sep=' $ '))
        return '( %s )' % ''.join(result)


class NameForm(SchemaElement):
    """
    Arguments:

    schema_element_str
      String containing an NameFormDescription

    Class attributes:

    oid
      OID of the name form
    names
      This list of strings contains all NAMEs of the name form
    desc
      This string contains description text (DESC) of the name form
    obsolete
      Integer flag (0 or 1) indicating whether the name form is marked
      as OBSOLETE in the schema
    form
      List of strings with NAMEs or OIDs of associated name forms
    oc
      String with NAME or OID of structural object classes this name form
      is usable with
    must
      This list of strings contains NAMEs or OIDs of all attributes
      an RDN must contain
    may
      This list of strings contains NAMEs or OIDs of additional attributes
      an RDN may contain
    """
    __slots__ = (
        'names',
        'oc',
        'must',
        'may',
    )
    schema_attribute = 'nameForms'
    token_defaults: TokenDefaults = {
        'NAME': (()),
        'DESC': (None,),
        'OBSOLETE': None,
        'OC': (None,),
        'MUST': (()),
        'MAY': (()),
    }

    def _set_attrs(self, opts):
        self.names = opts['NAME']
        self.desc = opts['DESC'][0]
        self.obsolete = opts['OBSOLETE'] is not None
        self.oc = opts['OC'][0]
        self.must = opts['MUST']
        self.may = opts['MAY']

    def __str__(self):
        result = [str(self.oid)]
        result.append(self.key_list('NAME', self.names, quoted=1))
        result.append(self.key_attr('DESC', self.desc, quoted=1))
        result.append({0:'', 1:' OBSOLETE'}[self.obsolete])
        result.append(self.key_attr('OC', self.oc))
        result.append(self.key_list('MUST', self.must, sep=' $ '))
        result.append(self.key_list('MAY', self.may, sep=' $ '))
        return '( %s )' % ''.join(result)


class SchemaElementOIDSet(MutableSet):
    """
    set-like class which handles OIDs and NAMEs of schema elements
    """

    def __init__(self, schema, se_class, nameoroids):
        self._schema = schema
        self._se_class = se_class
        self._nameoroid_dict = OrderedDict()
        self.update(nameoroids)

    def __iter__(self):
        for key in self._nameoroid_dict.keys():
            yield key

    def __len__(self):
        return len(self._nameoroid_dict)

    def _key_oid(self, nameoroid):
        return self._schema.get_oid(self._se_class, nameoroid, raise_keyerror=False).lower()

    def discard(self, nameoroid):
        try:
            del self._nameoroid_dict[self._key_oid(nameoroid)]
        except KeyError:
            pass

    def __contains__(self, nameoroid):
        return self._key_oid(nameoroid) in self._nameoroid_dict

    def intersection(self, s):
        return SchemaElementOIDSet(
            self._schema,
            self._se_class,
            [
                i
                for i in s
                if i in self
            ]
        )

    def add(self, se_name):
        se_name = se_name.strip()
        if not se_name:
            return
        if se_name[0] == '@':
            if self._se_class != AttributeType:
                raise ValueError('@-form only possible with class AttributeType')
            must_attr, may_attr = self._schema.attribute_types(
                (se_name[1:],),
                attr_type_filter=None,
                raise_keyerror=True,
                ignore_dit_content_rule=True
            )
            se_list = [
                (se_oid, (se_obj.names or [se_oid])[0])
                for se_oid, se_obj in list(must_attr.items()) + list(may_attr.items())
            ]
        else:
            se_list = [(self._key_oid(se_name), se_name)]
        for se_oid, se_name0 in se_list:
            if not se_oid in self._nameoroid_dict:
                self._nameoroid_dict[se_oid] = se_name0
        # end of add()

    def update(self, seq):
        for item in seq:
            self.add(item)

    @property
    def names(self):
        return list(self._nameoroid_dict.values())


class Entry(UserDict):
    """
    Schema-aware implementation of an LDAP entry class.

    Mainly it holds the attributes in a string-keyed dictionary with
    the OID as key.
    """

    def __init__(self, schema, dn, entry):
        self._keytuple2attrtype = {}
        self._attrtype2keytuple = {}
        self._s = schema
        self.dn = dn
        UserDict.__init__(self, {})
        self.update(entry)

    def name2key(self, nameoroid):
        """
        Return tuple of OID and all sub-types of attribute type specified
        in nameoroid.
        """
        if isinstance(nameoroid, tuple):
            return nameoroid
        try:
            # Mapping already in cache
            return self._attrtype2keytuple[nameoroid]
        except KeyError:
            # Mapping has to be constructed
            noid_l = nameoroid.lower().split(';')
            oid = self._s.get_oid(AttributeType, noid_l[0])
            if oid.lower() == 'entrydn':
                oid = '1.3.6.1.1.20'
            self._attrtype2keytuple[nameoroid] = (oid,)+tuple(noid_l[1:])
            return self._attrtype2keytuple[nameoroid]

    def __contains__(self, nameoroid):
        return self.name2key(nameoroid) in self.data

    def __getitem__(self, nameoroid):
        key = self.name2key(nameoroid)
        try:
            return self.data[key]
        except KeyError as err:
            if (
                    key[0] == '1.3.6.1.1.20' and
                    self.dn is not None
                ):
                return [self.dn.encode('utf-8')]
            raise err

    def __setitem__(self, nameoroid, attr_values):
        if isinstance(nameoroid, bytes):
            nameoroid = nameoroid.decode('ascii')
        key = self.name2key(nameoroid)
        self._keytuple2attrtype[key] = nameoroid
        self.data[key] = attr_values

    def __delitem__(self, nameoroid):
        key = self.name2key(nameoroid)
        del self._keytuple2attrtype[key]
        del self.data[key]

    def __iter__(self):
        for key in self.data.keys():
            yield self._keytuple2attrtype[key]

    def has_key(self, key):
        raise NotImplementedError('do not use %s.has_key() method' % (self.__class__.__name__,))

    def keys(self):
        return self.__iter__()

    def items(self):
        for k in self.keys():
            yield (k, self[k])

    def attribute_types(
            self,
            attr_type_filter=None,
            raise_keyerror=1
        ):
        """
        Convenience wrapper around SubSchema.attribute_types() which
        passes object classes of this particular entry as argument to
        SubSchema.attribute_types()
        """
        return self._s.attribute_types(
            self.object_class_oid_set(),
            attr_type_filter,
            raise_keyerror,
        )

    def object_class_oid_set(self):
        try:
            object_classes = [
                oc.decode('ascii')
                for oc in self.__getitem__('objectClass')
            ]
        except KeyError:
            object_classes = []
        return SchemaElementOIDSet(self._s, ObjectClass, object_classes)

    def get_structural_oc(self):
        try:
            structural_object_class_oid = self._s.get_oid(
                ObjectClass,
                self.__getitem__('structuralObjectClass')[-1].decode('ascii')
            )
        except (KeyError, IndexError):
            try:
                structural_object_class_oid = self._s.get_structural_oc(
                    [av.decode('ascii') for av in self.__getitem__('objectClass')]
                )
            except KeyError:
                return None
        return structural_object_class_oid

    def get_possible_dit_structure_rules(self, dn):
        try:
            structural_oc = self.get_structural_oc()
        except KeyError:
            return None
        else:
            return self._s.get_possible_dit_structure_rules(dn, structural_oc)

    def get_rdn_templates(self):
        return self._s.get_rdn_templates(self.get_structural_oc())



# -*- coding: ascii -*-
"""
ldap0.cidict - custom dictionary classes with case-insensitive string-keys
"""

from collections import UserDict


class CIDict(UserDict):
    """
    Case-insensitive but case-respecting dictionary.
    """

    def __init__(self, default=None, keytype=None):
        self._keys = {}
        self._keytype = keytype
        UserDict.__init__(self, {})
        self.update(default or {})

    def __getitem__(self, key):
        return self.data[key.lower()]

    def __setitem__(self, key, value):
        if self._keytype is None:
            self._keytype = type(key)
        if not isinstance(key, self._keytype):
            raise TypeError('Expected %s for key, got %r' % (self._keytype.__name__, key))
        lower_key = key.lower()
        self._keys[lower_key] = key
        self.data[lower_key] = value

    def __delitem__(self, key):
        lower_key = key.lower()
        del self._keys[lower_key]
        del self.data[lower_key]

    def update(self, data):
        for key in data.keys():
            self[key] = data[key]

    def has_key(self, key):
        raise NotImplementedError('do not use %s.has_key() method' % (self.__class__.__name__,))

    def __contains__(self, key):
        return UserDict.__contains__(self, key.lower())

    def keys(self):
        return self._keys.values()

    def items(self):
        for k in self._keys.values():
            yield (k, self[k])


class SingleValueDict(CIDict):
    """
    dictionary for reading only the 1st attribute value
    (strictly relational data model)
    """

    def __setitem__(self, key, val):
        return CIDict.__setitem__(self, key, val[0])



# -*- coding: ascii -*-
"""
ldap0.tmpl - templating
"""

from typing import Optional, Tuple
from collections import UserDict
from io import BytesIO

from .typehints import EntryMixed, EntryStr
from .base import decode_entry_dict
from .dn import escape_str as escape_dn_str
from .ldif import LDIFParser


class EscapeStrDict(UserDict):
    """
    dict drop-in replacement which applies a function
    before returning a dict value
    """

    def __init__(self, func, *args, **kwargs):
        UserDict.__init__(self, *args, **kwargs)
        self._func = func

    def __getitem__(self, key) -> str:
        return self._func(str(UserDict.__getitem__(self, key)))


class TemplateEntry:
    """
    Base class typically not directly used
    """
    t_dn: Optional[str] = None
    t_entry: Optional[EntryStr] = None

    def __init__(self, t_dn: str, t_entry: EntryStr):
        self.t_dn = t_dn
        self.t_entry = t_entry

    @classmethod
    def from_ldif_file(cls, fileobj, encoding: str = 'utf-8'):
        """
        create instance by reading LDIF from file-like object
        """
        t_dn, t_entry = LDIFParser(fileobj).list_entry_records()[0]
        return cls(
            t_dn.decode(encoding),
            decode_entry_dict(t_entry, encoding=encoding),
        )

    @classmethod
    def from_ldif_str(cls, buf):
        return cls.from_ldif_file(BytesIO(buf))

    def ldap_entry(
            self,
            data_dict,
            encoding: str = 'utf-8',
            first_stop_attrs = None,
        ) -> Tuple[str, EntryMixed]:
        """
        returns 2-tuple (dn, entry) derived from
        string-keyed, single-valued :data_dict:
        """
        ldap_entry = {}
        first_stop_attrs = set(first_stop_attrs or [])
        for attr_type in self.t_entry:
            attr_values = []
            attr_value_set = set()
            for av_template in self.t_entry[attr_type]:
                try:
                    attr_value = av_template.format(**data_dict).encode(encoding)
                except KeyError:
                    continue
                if attr_value and attr_value not in attr_value_set:
                    attr_values.append(attr_value)
                    attr_value_set.add(attr_value)
                if attr_type in first_stop_attrs:
                    break
            if attr_values:
                ldap_entry[attr_type] = attr_values
        dn_data_dict = EscapeStrDict(escape_dn_str, data_dict)
        ldap_dn = self.t_dn.format(**dn_data_dict)
        return (ldap_dn, ldap_entry)
